from perturbationx.toponpa.preprocessing import preprocessing, preprocess_network

from .preprocessing import *

__all__ = ['preprocessing', 'preprocess_network', 'format_dataset', 'prune_network_dataset', 'infer_graph_attributes']



import logging
import warnings
import itertools
from typing import Optional

import networkx as nx

from perturbationx.io import RelationTranslator


def infer_node_type(graph: nx.DiGraph):
    """Infer the type of each node in the network (core or boundary).

    :param graph: The network to process.
    :type graph: nx.DiGraph
    :raises ValueError: If the same node appears in both the core and boundary sets.
    :return: A tuple with the sets of boundary and core nodes.
    :rtype: (set, set)
    """
    # Select nodes with outgoing edges and targets of core edges as core nodes
    core_nodes = {src for src, trg in graph.edges} | \
                 {trg for src, trg in graph.edges if graph[src][trg]["type"] == "core"}

    # Select targets of boundary edges as boundary nodes
    boundary_nodes = {trg for src, trg in graph.edges if graph[src][trg]["type"] == "boundary"}

    # Check that there isn't overlap between the two sets
    node_intersection = core_nodes & boundary_nodes
    if len(node_intersection) > 0:
        raise ValueError("The same nodes appear in network core and boundary: %s." % str(node_intersection))

    # Infer targets of unknown edges; their type defaults to boundary if they have no outgoing links
    for src, trg in graph.edges:
        if graph[src][trg]["type"] not in ("core", "boundary"):
            if trg in core_nodes:
                graph[src][trg]["type"] = "core"
            else:
                graph[src][trg]["type"] = "boundary"
                boundary_nodes.add(trg)

    return boundary_nodes, core_nodes


def enumerate_nodes(graph: nx.DiGraph, boundary_nodes: set, core_nodes: set):
    """Assign an index and type to each node in the network.

    :param graph: The network to process.
    :type graph: nx.DiGraph
    :param boundary_nodes: The set of boundary nodes.
    :type boundary_nodes: set
    :param core_nodes: The set of core nodes.
    :type core_nodes: set
    """
    core_size = len(core_nodes)
    node_idx = {node: idx for idx, node in enumerate(core_nodes)} | \
               {node: (core_size + idx) for idx, node in enumerate(boundary_nodes)}
    for n in graph.nodes:
        graph.nodes[n]["idx"] = node_idx[n]
        graph.nodes[n]["type"] = "core" if n in core_nodes else "boundary"


def remove_invalid_graph_elements(graph: nx.DiGraph):
    """Remove invalid elements from the graph. This function removes self-loops and opposing edges, and ensures that the
    core graph is weakly connected.

    :param graph: The network to process.
    :type graph: nx.DiGraph
    """
    core_graph = graph.subgraph([n for n in graph.nodes if graph.nodes[n]["type"] == "core"])

    # Check that the core graph is weakly connected
    if not nx.is_weakly_connected(core_graph):
        warnings.warn("The network core is not weakly connected. Automatically selecting the largest component.")
        weak_components = sorted(nx.weakly_connected_components(core_graph), key=len, reverse=True)
        core_nodes_to_remove = set(itertools.chain.from_iterable(weak_components[1:]))
        removed_edge_count = sum(1 for src, trg in graph.edges
                                 if src in core_nodes_to_remove
                                 or trg in core_nodes_to_remove)
        graph.remove_nodes_from(core_nodes_to_remove)

        # Some boundary nodes may have become isolated
        boundary_nodes_to_remove = set(nx.isolates(graph))
        graph.remove_nodes_from(boundary_nodes_to_remove)
        warnings.warn("Removed %d nodes and %d associated edges: %s"
                      % (len(core_nodes_to_remove) + len(boundary_nodes_to_remove),
                         removed_edge_count, str(core_nodes_to_remove.union(boundary_nodes_to_remove))))

    # Check that the graph has no self-loops
    self_loops = list(nx.selfloop_edges(graph))
    if len(self_loops) > 0:
        warnings.warn("The network contains self-loops. "
                      "They cannot be processed by the algorithm and will be removed. "
                      "Found %d self-loops on nodes: %s" %
                      (len(self_loops), str(set(loop[0] for loop in self_loops))))
        graph.remove_edges_from(self_loops)

    # Check that the graph has no opposing edges
    opposing_edges = set((src, trg) for src, trg in graph.edges
                         if (trg, src) in graph.edges and src >= trg)
    if len(opposing_edges) > 0:
        warnings.warn("The network contains opposing edges. "
                      "They will be collapsed into a single edge with their weights added. "
                      "Found %d pairs of opposing edges: %s" % (len(opposing_edges), str(opposing_edges)))


def infer_edge_attributes(graph: nx.DiGraph, relation_translator: Optional[RelationTranslator] = None):
    """Infer the attributes of each edge in the network.

    :param graph: The network to process.
    :type graph: nx.DiGraph
    :param relation_translator: The relation translator to use. If None, a new instance will be created.
    :type relation_translator: perturbationx.RelationTranslator, optional
    """
    rt = relation_translator if relation_translator is not None else RelationTranslator()
    for src, trg in graph.edges:
        edge_weight = rt.translate(graph[src][trg]["relation"])
        graph[src][trg]["weight"] = edge_weight
        if edge_weight > 0:
            graph[src][trg]["interaction"] = "directlyIncreases"
        elif edge_weight < 0:
            graph[src][trg]["interaction"] = "directlyDecreases"
        else:
            graph[src][trg]["interaction"] = "causesNoChange"


def infer_metadata(graph: nx.DiGraph, verbose=True):
    """Infer metadata about the network and add it to the graph instance.

    :param graph: The network to process.
    :type graph: nx.DiGraph
    :param verbose: Whether to log network statistics.
    :type verbose: bool, optional
    """
    boundary_nodes = {trg for src, trg in graph.edges if graph[src][trg]["type"] == "boundary"}
    core_nodes = {src for src, trg in graph.edges} | \
                 {trg for src, trg in graph.edges if graph[src][trg]["type"] == "core"}
    inner_boundary_nodes = {src for src, trg in graph.edges if graph[src][trg]["type"] == "boundary"}

    core_edge_count = sum(1 for e in graph.edges.data() if e[2]["type"] == "core")
    boundary_edge_count = sum(1 for e in graph.edges.data() if e[2]["type"] == "boundary")

    graph.graph["core_edges"] = core_edge_count
    graph.graph["boundary_edges"] = boundary_edge_count
    graph.graph["core_nodes"] = len(core_nodes)
    graph.graph["outer_boundary_nodes"] = len(boundary_nodes)
    graph.graph["inner_boundary_nodes"] = len(inner_boundary_nodes)

    if verbose:  # Log network statistics
        logging.info("core edges: %d" % core_edge_count)
        logging.info("boundary edges: %d" % boundary_edge_count)
        logging.info("core nodes: %d" % len(core_nodes))
        logging.info("outer boundary nodes: %d" % len(boundary_nodes))
        logging.info("core nodes with boundary edges (inner boundary): %d" % len(inner_boundary_nodes))




import numpy as np
import numpy.linalg as la
from scipy.sparse import issparse, sparray

__all__ = ["coefficient_inference", "perturbation_amplitude", "perturbation_amplitude_contributions"]


def coefficient_inference(lap_b: np.ndarray | sparray, lap_c: np.ndarray | sparray, boundary_coefficients: np.ndarray):
    """Infer core coefficients from boundary coefficients and Laplacian matrices.

    :param lap_b: The Lb boundary Laplacian.
    :type lap_b: np.ndarray | sp.sparray
    :param lap_c: The Lc core Laplacian.
    :type lap_c: np.ndarray | sp.sparray
    :param boundary_coefficients: The boundary coefficients.
    :type boundary_coefficients: np.ndarray
    :raises ValueError: If the Laplacian matrices are misshapen or if the matrix dimensions do not match.
    :return: The inferred core coefficients.
    :rtype: np.ndarray
    """
    if lap_b.ndim != 2:
        raise ValueError("Argument lap_b is not two-dimensional.")
    elif lap_c.ndim != 2 or lap_c.shape[0] != lap_c.shape[1]:
        raise ValueError("Argument lap_c not a square matrix.")
    elif lap_c.shape[0] != lap_b.shape[0]:
        raise ValueError("Dimensions of lap_b and lap_c do not match.")
    elif lap_b.shape[1] != boundary_coefficients.shape[0]:
        raise ValueError("Dimensions of lap_b and boundary_coefficients do not match.")

    if issparse(lap_c):
        lap_c = lap_c.todense()

    edge_constraints = - lap_b @ boundary_coefficients
    return la.solve(lap_c, edge_constraints)


def perturbation_amplitude(lap_q: np.ndarray | sparray, core_coefficients: np.ndarray, core_edge_count: int):
    """Compute the perturbation amplitude from the core Laplacian and core coefficients.

    :param lap_q: The Q core Laplacian.
    :type lap_q: np.ndarray | sp.sparray
    :param core_coefficients: The core coefficients.
    :type core_coefficients: np.ndarray
    :param core_edge_count: The number of edges in the core network.
    :type core_edge_count: int
    :raises ValueError: If the Laplacian matrix is misshapen or if the matrix dimensions do not match.
    :return: The perturbation amplitude.
    :rtype: np.ndarray
    """
    if lap_q.ndim != 2 or lap_q.shape[0] != lap_q.shape[1]:
        raise ValueError("Argument lap_q is not a square matrix.")
    elif lap_q.shape[0] != core_coefficients.shape[0]:
        raise ValueError("Dimensions of lap_q and core_coefficients do not match.")

    total_amplitude = core_coefficients.T @ lap_q @ core_coefficients
    if core_coefficients.ndim > 1 and core_coefficients.shape[1] > 1:
        total_amplitude = np.diag(total_amplitude)
    return total_amplitude / core_edge_count


def perturbation_amplitude_contributions(lap_q: np.ndarray | sparray, core_coefficients: np.ndarray,
                                         core_edge_count: int):
    """Compute the perturbation amplitude and relative contributions from the core Laplacian and core coefficients.

    :param lap_q: The Q core Laplacian.
    :type lap_q: np.ndarray | sp.sparray
    :param core_coefficients: The core coefficients.
    :type core_coefficients: np.ndarray
    :param core_edge_count: The number of edges in the core network.
    :type core_edge_count: int
    :raises ValueError: If the Laplacian matrix is misshapen or if the matrix dimensions do not match.
    :return: The perturbation amplitude and relative contributions.
    :rtype: (np.ndarray, np.ndarray)
    """
    if lap_q.ndim != 2 or lap_q.shape[0] != lap_q.shape[1]:
        raise ValueError("Argument lap_q is not a square matrix.")
    elif lap_q.shape[0] != core_coefficients.shape[0]:
        raise ValueError("Dimensions of lap_q and core_coefficients do not match.")

    node_contributions = np.multiply(lap_q @ core_coefficients, core_coefficients)
    total_perturbation = node_contributions.sum(axis=0)

    relative_contributions = np.divide(node_contributions, total_perturbation)
    relative_contributions[relative_contributions == -0.] = 0.
    return total_perturbation / core_edge_count, relative_contributions.T



import warnings
import logging
from pathlib import Path
from typing import Optional
import json

import numpy as np
import pandas as pd
import networkx as nx

import perturbationx.io as px_io
from perturbationx.io import RelationTranslator
import perturbationx.toponpa as toponpa
from perturbationx.resources import DEFAULT_DATA_COLS, DEFAULT_LOGGING_KWARGS


class CausalNetwork:
    """Class for storing and evaluating causal networks. The network is stored as a networkx.DiGraph. The class
    provides methods for reading and writing networks from and to various file formats, as well as for evaluating
    the network using the TopoNPA algorithm. The class also provides methods for rewiring and wiring edges in the
    network, as well as for evaluating the rewired and wired networks.

    :param graph: The network graph. Defaults to None (empty graph).
    :type graph: networkx.DiGraph, optional
    :param relation_translator: The relation translator. Defaults to None (default relation translator).
    :type relation_translator: perturbationx.RelationTranslator, optional
    :param inplace: Whether to initialize the network in place. If False, the network is copied. Defaults to False.
    :type inplace: bool, optional
    :raises TypeError: If the relation translator is not a RelationTranslator.
    """
    __allowed_edge_types = ("core", "boundary", "infer")

    def __init__(self, graph: Optional[nx.DiGraph] = None,
                 relation_translator: Optional[RelationTranslator] = None,
                 inplace=False):
        if relation_translator is None:
            relation_translator = RelationTranslator()
        if not isinstance(relation_translator, RelationTranslator):
            raise TypeError("Argument relation_translator is not a RelationTranslator.")
        self.relation_translator = relation_translator

        if graph is None:
            graph = nx.DiGraph()
        elif not isinstance(graph, nx.DiGraph):
            warnings.warn("Argument graph is not a networkx.DiGraph, and will be converted.")
            graph = nx.DiGraph(graph)
        elif not inplace:
            graph = graph.copy()

        px_io.validate_nx_graph(graph, self.__allowed_edge_types)
        self._graph = graph

        # Copy metadata from nx attributes, clear attributes and set default values
        self.metadata = self._graph.graph.copy()
        self._graph.graph.clear()
        self.initialize_metadata()

        self._cytoscape_suid = dict()

    @classmethod
    def from_networkx(cls, graph: nx.DiGraph,
                      relation_translator: Optional[RelationTranslator] = None,
                      inplace=False):
        """Construct a new CausalNetwork from a networkx.DiGraph.

        :param graph: The network graph.
        :type graph: networkx.DiGraph
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :param inplace: Whether to initialize the network in place. If False, the network is copied. Defaults to False.
        :type inplace: bool, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        """
        return cls(graph, relation_translator, inplace)

    @classmethod
    def from_dsv(cls, filepath: str, edge_type="infer", delimiter='\t',
                 header_cols=DEFAULT_DATA_COLS, relation_translator=None):
        """Construct a new CausalNetwork from a delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param delimiter: The delimiter to use. Defaults to '\t'.
        :type delimiter: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        :return: The constructed CausalNetwork.
        :rtype: CausalNetwork
        """
        graph = nx.DiGraph()
        graph.add_edges_from(px_io.read_dsv(
            filepath, default_edge_type=edge_type, delimiter=delimiter, header_cols=header_cols
        ))
        graph.graph["title"] = Path(filepath).stem
        graph.graph["collection"] = Path(filepath).parent.name
        return cls(graph, relation_translator, inplace=True)

    @classmethod
    def from_tsv(cls, filepath: str, edge_type="infer",
                 header_cols=DEFAULT_DATA_COLS, relation_translator=None):
        """Construct a new CausalNetwork from a tab-delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        :return: The constructed CausalNetwork.
        :rtype: CausalNetwork
        """
        return cls.from_dsv(filepath, edge_type, header_cols=header_cols,
                            relation_translator=relation_translator)

    @classmethod
    def from_csv(cls, filepath: str, edge_type="infer",
                 header_cols=DEFAULT_DATA_COLS, relation_translator=None):
        """Construct a new CausalNetwork from a comma-delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        :return: The constructed CausalNetwork.
        :rtype: CausalNetwork
        """
        return cls.from_dsv(filepath, edge_type, delimiter=',', header_cols=header_cols,
                            relation_translator=relation_translator)

    @classmethod
    def from_cyjs_json(cls, filepath: str, relation_translator=None):
        """Construct a new CausalNetwork from a Cytoscape.js JSON file.

        :param filepath: The path to the file.
        :type filepath: str
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        :return: The constructed CausalNetwork.
        :rtype: CausalNetwork
        """
        with open(filepath) as f:
            graph = nx.cytoscape_graph(json.load(f))
            graph.graph["title"] = Path(filepath).stem
            graph.graph["collection"] = Path(filepath).parent.name
            return cls(graph, relation_translator, inplace=True)

    @classmethod
    def from_pandas(cls, df: pd.DataFrame, default_edge_type="infer",
                    header_cols=DEFAULT_DATA_COLS, relation_translator=None):
        """Construct a new CausalNetwork from a pandas DataFrame.

        :param df: The DataFrame to parse.
        :type df: pd.DataFrame
        :param default_edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type default_edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        :param relation_translator: The relation translator. Defaults to None (default relation translator).
        :type relation_translator: perturbationx.RelationTranslator, optional
        :raises TypeError: If the relation translator is not a RelationTranslator.
        :return: The constructed CausalNetwork.
        :rtype: CausalNetwork
        """
        edge_list = px_io.parse_pandas(
            df, default_edge_type=default_edge_type, header_cols=header_cols
        )
        graph = nx.DiGraph()
        graph.add_edges_from(edge_list)
        return cls(graph, relation_translator, inplace=True)

    def add_edges_from_dsv(self, filepath: str, edge_type="infer", delimiter='\t',
                           header_cols=DEFAULT_DATA_COLS):
        """Add edges from a delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param delimiter: The delimiter to use. Defaults to '\t'.
        :type delimiter: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        """
        self._graph.add_edges_from(px_io.read_dsv(
            filepath, default_edge_type=edge_type, delimiter=delimiter, header_cols=header_cols
        ))
        px_io.validate_nx_graph(self._graph, self.__allowed_edge_types)

    def add_edges_from_tsv(self, filepath: str, edge_type="infer",
                           header_cols=DEFAULT_DATA_COLS):
        """Add edges from a tab-delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        """
        self.add_edges_from_dsv(filepath, edge_type, header_cols=header_cols)

    def add_edges_from_csv(self, filepath: str, edge_type="infer",
                           header_cols=DEFAULT_DATA_COLS):
        """Add edges from a comma-delimited file.

        :param filepath: The path to the file.
        :type filepath: str
        :param edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        """
        self.add_edges_from_dsv(filepath, edge_type, delimiter=',', header_cols=header_cols)

    def add_edges_from_pandas(self, df: pd.DataFrame, default_edge_type="infer",
                              header_cols=DEFAULT_DATA_COLS):
        """Add edges from a pandas DataFrame.

        :param df: The DataFrame to parse.
        :type df: pd.DataFrame
        :param default_edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
        :type default_edge_type: str, optional
        :param header_cols: The column names of the columns in the file. Defaults to ["subject", "object", "relation",
            "type"].
        :type header_cols: list, optional
        """
        self._graph.add_edges_from(px_io.parse_pandas(
            df, default_edge_type=default_edge_type, header_cols=header_cols
        ))
        px_io.validate_nx_graph(self._graph, self.__allowed_edge_types)

    def initialize_metadata(self):
        """Initialize the metadata dictionary with default values. These values are "Untitled network" for the title
        and "Untitled collection" for the collection. The metadata dictionary is created if it does not exist.
        """
        # The user can delete metadata dict
        if self.metadata is None:
            self.metadata = dict()
        # Title and collection are required for Cytoscape
        if "title" not in self.metadata:
            self.metadata["title"] = "Untitled network"
        if "collection" not in self.metadata:
            self.metadata["collection"] = "Untitled collection"

    def copy(self):
        """Return a copy of the CausalNetwork.

        :return: The copy.
        :rtype: CausalNetwork
        """
        return CausalNetwork(self._graph, self.relation_translator.copy(), inplace=False)

    def number_of_nodes(self, typ=None):
        """Return the number of nodes in the network.

        :param typ: The node type to count. If None, all nodes are counted. Defaults to None.
        :type typ: str, optional
        :return: The number of nodes.
        :rtype: int
        """
        if typ is not None:
            return sum(1 for n, d in self._graph.nodes(data=True) if "type" in d and d["type"] == typ)
        return self._graph.number_of_nodes()

    def nodes(self, typ=None, data=True):
        """Return a list of nodes in the network.

        :param typ: The node type to return. If None, all nodes are returned. Defaults to None.
        :type typ: str, optional
        :param data: Whether to return the node data. Defaults to True.
        :type data: bool, optional
        :return: The list of nodes.
        :rtype: list
        """
        if typ is not None:
            if data:
                return [(n, d) for n, d in self._graph.nodes(data=True) if "type" in d and d["type"] == typ]
            else:
                return [n for n, d in self._graph.nodes(data=True) if "type" in d and d["type"] == typ]

        return list(self._graph.nodes(data=data))

    def number_of_edges(self, typ=None):
        """Return the number of edges in the network.

        :param typ: The edge type to count. If None, all edges are counted. Defaults to None.
        :type typ: str, optional
        :return: The number of edges.
        :rtype: int
        """
        if typ is not None:
            return sum(1 for e in self._graph.edges.data() if "type" in e[2] and e[2]["type"] == typ)
        return self._graph.number_of_edges()

    def edges(self, typ=None, data=True):
        """Return a list of edges in the network.

        :param typ: The edge type to return. If None, all edges are returned. Defaults to None.
        :type typ: str, optional
        :param data: Whether to return the edge data. Defaults to True.
        :type data: bool, optional
        :return: The list of edges.
        :rtype: list
        """
        if typ is not None:
            if data:
                return [(src, trg, d) for src, trg, d in self._graph.edges.data() if "type" in d and d["type"] == typ]
            else:
                return [(src, trg) for src, trg, d in self._graph.edges.data() if "type" in d and d["type"] == typ]

        if data:
            return list(self._graph.edges.data())
        else:
            return list(self._graph.edges)

    def add_edge(self, src: str, trg: str, rel: str, typ="infer", confidence=None):
        """Add an edge to the network. If the edge already exists, it is modified. If the nodes do not exist, they are
        created.

        :param src: The source node.
        :type src: str
        :param trg: The target node.
        :type trg: str
        :param rel: The causal relation of the edge.
        :type rel: str
        :param typ: The type of the edge. Allowed values are "core", "boundary" and "infer". Defaults to "infer".
        :type typ: str, optional
        :param confidence: The confidence of the edge. Defaults to None, which sets the confidence to 1.0.
        :type confidence: float, optional
        """
        if self._graph.has_edge(src, trg):
            warnings.warn("Edge between %s and %s already exists "
                          "and will be modified." % (src, trg))
        if typ in self.__allowed_edge_types:
            self._graph.add_edge(src, trg, relation=rel, type=typ)
        else:
            warnings.warn("Unknown type %s of edge %s will be "
                          "replaced with \"infer\"." % (typ, str((src, trg))))
            self._graph.add_edge(src, trg, relation=rel, type="infer")
        if confidence is not None:
            self._graph[src][trg]["confidence"] = float(confidence)

    def modify_edge(self, src: str, trg: str, rel=None, typ=None, confidence=None):
        """Modify an edge in the network. If the edge does not exist, a KeyError is raised.

        :param src: The source node.
        :type src: str
        :param trg: The target node.
        :type trg: str
        :param rel: The new causal relation of the edge. Defaults to None, which does not modify the relation.
        :type rel: str, optional
        :param typ: The new type of the edge. Allowed values are "core", "boundary" and "infer". Defaults to None,
            which does not modify the type.
        :type typ: str, optional
        :param confidence: The new confidence of the edge. Defaults to None, which does not modify the confidence.
        :type confidence: float, optional
        :raises KeyError: If the edge does not exist.
        """
        if not self._graph.has_edge(src, trg):
            raise KeyError("Edge between %s and %s does not exist." % (src, trg))
        if rel is not None:
            self._graph[src][trg]["relation"] = rel
        if typ is not None:
            if typ in self.__allowed_edge_types:
                self._graph[src][trg]["type"] = typ
            else:
                warnings.warn("Unknown type %s of edge %s "
                              "will be ignored." % (typ, str((src, trg))))
        if confidence is not None:
            self._graph[src][trg]["confidence"] = float(confidence)

    def remove_edge(self, src: str, trg: str):
        """Remove an edge from the network. If the edge does not exist, a KeyError is raised.

        :param src: The source node.
        :type src: str
        :param trg: The target node.
        :type trg: str
        :raises KeyError: If the edge does not exist.
        """
        if not self._graph.has_edge(src, trg):
            raise KeyError("Edge between %s and %s does not exist." % (src, trg))
        self._graph.remove_edge(src, trg)

        if self._graph.degree[src] == 0:
            self._graph.remove_node(src)
        if self._graph.degree[trg] == 0:
            self._graph.remove_node(trg)

    def modify_network(self, edge_list):
        """Modify the network using a list of edges. The list of edges is a list of tuples of the form
        (source, target, relation, type). If the relation is None, the edge is removed. If the type is None, the type
        is not modified. If the type is not one of the allowed types, it is ignored or replaced with "infer".

        :param edge_list: The list of edges.
        :type edge_list: list
        """
        for src, trg, rel, typ in edge_list:
            if rel is None:
                self.remove_edge(src, trg)
            elif typ in self.__allowed_edge_types:
                self._graph.add_edge(src, trg, relation=rel, type=typ)
            elif self._graph.has_edge(src, trg):
                warnings.warn("Unknown type %s of edge %s "
                              "will be ignored." % (typ, str((src, trg))))
                self._graph[src][trg]["relation"] = rel
            else:
                warnings.warn("Unknown type %s of edge %s will be "
                              "replaced with \"infer\"." % (typ, str((src, trg))))
                self._graph.add_edge(src, trg, relation=rel, type="infer")

    def rewire_edges(self, nodes: list, iterations: int, datasets: list, method='k1', p_rate=1.,
                     missing_value_pruning_mode="nullify", opposing_value_pruning_mode=None,
                     opposing_value_minimum_amplitude=1., boundary_edge_minimum=6,
                     exact_boundary_outdegree=True, sparse=True, seed=None,
                     verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Rewire edges in the network. The rewiring is performed on the given nodes, and the given number of
        iterations. Datasets can be provided to evaluate the rewired networks. If no datasets are provided, the
        rewired networks are returned as a list of modifications. Otherwise, the NPAs of the rewired networks are
        computed and returned.

        :param nodes: The nodes to rewire.
        :type nodes: list
        :param iterations: The number of rewiring iterations.
        :type iterations: int
        :param datasets: The datasets to use for evaluation. If None, the rewired networks are returned as a list of
            modifications. Otherwise, the NPAs of the rewired networks are computed and returned.
        :type datasets: list
        :param method: The rewiring method to use. May be "k1" or "k2". Defaults to "k1".
        :type method: str, optional
        :param p_rate: The fraction of edges to rewire. Defaults to 1.0.
        :type p_rate: float, optional
        :param missing_value_pruning_mode: The mode to use for pruning nodes with missing values. Must be one of
            'remove' or 'nullify'. Defaults to 'nullify'.
        :type missing_value_pruning_mode: str, optional
        :param opposing_value_pruning_mode: The mode to use for pruning edges with opposing values. Must be one of
            'remove', 'nullify', or 'none'. Defaults to None.
        :type opposing_value_pruning_mode: str, optional
        :param opposing_value_minimum_amplitude: The minimum amplitude of the dataset values to consider. Values with
            an absolute value smaller than this threshold are ignored. Defaults to 1.
        :type opposing_value_minimum_amplitude: float, optional
        :param boundary_edge_minimum: The minimum number of boundary edges a core node must have to be included
            in the pruned network. If a core node has fewer boundary edges after 'remove' pruning, all of its edges are
            removed. This parameter is ignored if 'nullify' pruning is used. Defaults to 6.
        :type boundary_edge_minimum: int, optional
        :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
            is set to 1 for all core nodes with boundary edges. Defaults to True.
        :type exact_boundary_outdegree: bool, optional
        :param sparse: Whether to use sparse computation. Defaults to True.
        :type sparse: bool, optional
        :param seed: The seed to use for the random number generator. Defaults to None.
        :type seed: int, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the permutation method is unknown. If the same node appears in both
            the core and boundary network.
        :return: The list of modifications, or a list of tuples of the form (modification, npa), where modification
            is the modification and npa is a dictionary of the form {dataset_id: npa}.
        :rtype: list
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)
            logging.info("REWIRING EDGES")

        # Find existing edges
        existing_edges = []
        for src, trg in self._graph.edges(nodes):
            if trg in nodes:
                confidence = self._graph[src][trg].get("confidence", 1.0)
                existing_edges.append((src, trg, self._graph[src][trg]["relation"], confidence))

        # Permute edges
        modifications = toponpa.permute_edge_list(
            np.array(existing_edges), nodes, iterations,
            method=method, permutation_rate=p_rate, seed=seed
        )
        for idx in range(len(modifications)):
            modifications[idx] = [(*edge, "core") for edge in modifications[idx]]

        # If no datasets were given, return the permutations
        if datasets is None:
            return modifications

        # Otherwise, compute NPAs for each dataset
        return self.evaluate_modifications(
            modifications, nodes, datasets,
            missing_value_pruning_mode=missing_value_pruning_mode,
            opposing_value_pruning_mode=opposing_value_pruning_mode,
            opposing_value_minimum_amplitude=opposing_value_minimum_amplitude,
            boundary_edge_minimum=boundary_edge_minimum,
            exact_boundary_outdegree=exact_boundary_outdegree,
            sparse=sparse, seed=seed, verbose=verbose
        )

    def wire_edges(self, nodes: list, iterations: int, datasets: list, number_of_edges: int, edge_relations: list,
                   missing_value_pruning_mode="nullify", opposing_value_pruning_mode=None,
                   opposing_value_minimum_amplitude=1., boundary_edge_minimum=6,
                   exact_boundary_outdegree=True, sparse=True, seed=None,
                   verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Wire edges in the network. The wiring is performed on the given nodes, and the given number of
        iterations. Datasets can be provided to evaluate the wired networks. If no datasets are provided, the
        wired networks are returned as a list of modifications. Otherwise, the NPAs of the wired networks are
        computed and returned.

        :param nodes: The nodes to wire.
        :type nodes: list
        :param iterations: The number of wiring iterations.
        :type iterations: int
        :param datasets: The datasets to use for evaluation. If None, the wired networks are returned as a list of
            modifications. Otherwise, the NPAs of the wired networks are computed and returned.
        :type datasets: list
        :param number_of_edges: The number of edges to wire.
        :type number_of_edges: int
        :param edge_relations: The relations to use for wiring.
        :type edge_relations: list
        :param missing_value_pruning_mode: The mode to use for pruning nodes with missing values. Must be one of
            'remove' or 'nullify'. Defaults to 'nullify'.
        :type missing_value_pruning_mode: str, optional
        :param opposing_value_pruning_mode: The mode to use for pruning edges with opposing values. Must be one of
            'remove', 'nullify', or 'none'. Defaults to None.
        :type opposing_value_pruning_mode: str, optional
        :param opposing_value_minimum_amplitude: The minimum amplitude of the dataset values to consider. Values with
            an absolute value smaller than this threshold are ignored. Defaults to 1.
        :type opposing_value_minimum_amplitude: float, optional
        :param boundary_edge_minimum: The minimum number of boundary edges a core node must have to be included
            in the pruned network. If a core node has fewer boundary edges after 'remove' pruning, all of its edges are
            removed. This parameter is ignored if 'nullify' pruning is used. Defaults to 6.
        :type boundary_edge_minimum: int, optional
        :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
            is set to 1 for all core nodes with boundary edges. Defaults to True.
        :type exact_boundary_outdegree: bool, optional
        :param sparse: Whether to use sparse computation. Defaults to True.
        :type sparse: bool, optional
        :param seed: The seed to use for the random number generator. Defaults to None.
        :type seed: int, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The list of modifications, or a list of tuples of the form (modification, npa), where modification
            is the modification and npa is a dictionary of the form {dataset_id: npa}.
        :rtype: list
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)
            logging.info("WIRING EDGES")

        rng = np.random.default_rng(seed)
        modifications = []
        for _ in range(iterations):
            src_nodes = rng.choice(nodes, size=number_of_edges, replace=True)
            trg_nodes = rng.choice(nodes, size=number_of_edges, replace=True)
            for idx, (src, trg) in enumerate(zip(src_nodes, trg_nodes)):
                while trg == src:
                    trg = rng.choice(nodes)
                trg_nodes[idx] = trg

            relations = rng.choice(edge_relations, size=number_of_edges, replace=True)
            modifications.append([
                (src, trg, rel, "core") for src, trg, rel
                in zip(src_nodes, trg_nodes, relations)
            ])

        # If no datasets were given, return the modifications
        if datasets is None:
            return modifications

        # Otherwise, compute NPAs for each dataset
        return self.evaluate_modifications(
            modifications, nodes, datasets,
            missing_value_pruning_mode=missing_value_pruning_mode,
            opposing_value_pruning_mode=opposing_value_pruning_mode,
            opposing_value_minimum_amplitude=opposing_value_minimum_amplitude,
            boundary_edge_minimum=boundary_edge_minimum,
            exact_boundary_outdegree=exact_boundary_outdegree,
            sparse=sparse, seed=seed, verbose=verbose
        )

    def evaluate_modifications(self, modifications: list, nodes, datasets, missing_value_pruning_mode="nullify",
                               opposing_value_pruning_mode=None, opposing_value_minimum_amplitude=1.,
                               boundary_edge_minimum=6, exact_boundary_outdegree=True, sparse=True,
                               seed=None, verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Evaluate modifications of the network. The modifications are evaluated using the given datasets. The
        modifications are returned as a list of tuples of the form (modification, npa), where modification is the
        modification and npa is a dictionary of the form {dataset_id: npa}. Modifications can be generated using
        the rewire_edges or wire_edges methods.

        :param modifications: The modifications to evaluate. The modifications are a list of lists of tuples of the
            form (source, target, relation, type).
        :type modifications: list
        :param nodes: The nodes used in the modifications.
        :type nodes: list
        :param datasets: The datasets to use for evaluation.
        :type datasets: list
        :param missing_value_pruning_mode: The mode to use for pruning nodes with missing values. Must be one of
            'remove' or 'nullify'. Defaults to 'nullify'.
        :type missing_value_pruning_mode: str, optional
        :param opposing_value_pruning_mode: The mode to use for pruning edges with opposing values. Must be one of
            'remove', 'nullify', or 'none'. Defaults to None.
        :type opposing_value_pruning_mode: str, optional
        :param opposing_value_minimum_amplitude: The minimum amplitude of the dataset values to consider. Values with
            an absolute value smaller than this threshold are ignored. Defaults to 1.
        :type opposing_value_minimum_amplitude: float, optional
        :param boundary_edge_minimum: The minimum number of boundary edges a core node must have to be included
            in the pruned network. If a core node has fewer boundary edges after 'remove' pruning, all of its edges are
            removed. This parameter is ignored if 'nullify' pruning is used. Defaults to 6.
        :type boundary_edge_minimum: int, optional
        :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
            is set to 1 for all core nodes with boundary edges. Defaults to True.
        :type exact_boundary_outdegree: bool, optional
        :param sparse: Whether to use sparse computation. Defaults to True.
        :type sparse: bool, optional
        :param seed: The seed to use for the random number generator. Defaults to None.
        :type seed: int, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The list of modifications, or a list of tuples of the form (modification, npa), where modification
            is the modification and npa is a dictionary of the form {dataset_id: npa}.
        :rtype: list
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)

        # Copy the graph
        prograph = self._graph.to_directed()

        return toponpa.evaluate_modifications(
            prograph, self.relation_translator, modifications, nodes, datasets,
            missing_value_pruning_mode=missing_value_pruning_mode,
            opposing_value_pruning_mode=opposing_value_pruning_mode,
            opposing_value_minimum_amplitude=opposing_value_minimum_amplitude,
            boundary_edge_minimum=boundary_edge_minimum,
            exact_boundary_outdegree=exact_boundary_outdegree,
            sparse=sparse, seed=seed, verbose=verbose
        )

    def infer_graph_attributes(self, inplace=False, verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Infer graph attributes required for NPA computation. The attributes are inferred from the relations in the
        network. The attributes are added to the graph as node and edge attributes. If inplace is True, the attributes
        are added to the graph and the graph is returned. Otherwise, the attributes are added to a copy of the graph
        and the copy is returned.

        :param inplace: Whether to add the attributes to the graph inplace. Defaults to False.
        :type inplace: bool, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to  "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The graph with the inferred attributes.
        :rtype: CausalNetwork
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)

        if inplace:
            toponpa.infer_graph_attributes(self._graph, self.relation_translator, verbose)
            return self
        else:
            graph = self._graph.to_directed()
            toponpa.infer_graph_attributes(graph, self.relation_translator, verbose)
            return CausalNetwork(graph, self.relation_translator, inplace=True)

    def get_adjacencies(self, sparse=False, verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Get the adjacency matrices of the network.

        :param sparse: Whether to use sparse matrices. Defaults to False.
        :type sparse: bool, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to  "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The boundary and core adjacency matrices and the node ordering.
        :rtype: (np.ndarray | sp.sparray, np.ndarray | sp.sparray, list)
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)
            logging.info("PREPROCESSING NETWORK")

        # Preprocess the graph
        prograph = self._graph.to_directed()
        toponpa.infer_graph_attributes(prograph, self.relation_translator, verbose)

        # Construct adjacency matrices
        adj_b, adj_c = toponpa.generate_adjacencies(prograph, sparse=sparse)

        # Determine node ordering
        node_ordering = [None] * len(prograph.nodes)
        for node in prograph.nodes:
            node_ordering[prograph.nodes[node]["idx"]] = node

        return adj_b, adj_c, node_ordering

    def get_laplacians(self, boundary_outdegree_minimum=6, exact_boundary_outdegree=True, sparse=False,
                       verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Get the Laplacian matrices of the network.

        :param boundary_outdegree_minimum: The minimum number of boundary edges a core node must have to be included
            in the boundary network. Defaults to 6.
        :type boundary_outdegree_minimum: int, optional
        :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
            is set to 1 for all core nodes with boundary edges. Defaults to True.
        :type exact_boundary_outdegree: bool, optional
        :param sparse: Whether to use sparse matrices. Defaults to False.
        :type sparse: bool, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to  "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The Lb boundary Laplacian, Lc core Laplacian and Q core Laplacian and the node ordering.
        :rtype: (np.ndarray | sp.sparray, np.ndarray | sp.sparray, np.ndarray | sp.sparray, list)
        """
        adj_b, adj_c, node_ordering = self.get_adjacencies(
            sparse=sparse, verbose=verbose, logging_kwargs=logging_kwargs)
        lap_b = - toponpa.generate_boundary_laplacian(
            adj_b, boundary_edge_minimum=boundary_outdegree_minimum
        )
        lap_c, lap_q, = toponpa.generate_core_laplacians(
            lap_b, adj_c, exact_boundary_outdegree=exact_boundary_outdegree
        )
        return lap_b, lap_c, lap_q, node_ordering

    def toponpa(self, datasets: dict, missing_value_pruning_mode="nullify", opposing_value_pruning_mode=None,
                opposing_value_minimum_amplitude=1., boundary_edge_minimum=6, exact_boundary_outdegree=True,
                compute_statistics=True, alpha=0.95, permutations=('o', 'k2'), full_core_permutation=True,
                p_iters=500, p_rate=1., sparse=True, seed=None, verbose=True, logging_kwargs=DEFAULT_LOGGING_KWARGS):
        """Compute the Network Perturbation Amplitude (NPA) for a given network and datasets.

        :param datasets: The datasets to use. The keys are the dataset IDs and the values are the datasets, which are
            pandas DataFrames.
        :type datasets: dict
        :param missing_value_pruning_mode: The mode to use for pruning nodes with missing values. Must be one of
            'remove' or 'nullify'. Defaults to 'nullify'.
        :type missing_value_pruning_mode: str, optional
        :param opposing_value_pruning_mode: The mode to use for pruning edges with opposing values. Must be one of
            'remove', 'nullify', or 'none'. Defaults to None.
        :type opposing_value_pruning_mode: str, optional
        :param opposing_value_minimum_amplitude: The minimum amplitude of the dataset values to consider. Values with an
            absolute value smaller than this threshold are ignored. Defaults to 1.
        :type opposing_value_minimum_amplitude: float, optional
        :param boundary_edge_minimum: The minimum number of boundary edges a core node must have to be included in
            the pruned network. If a core node has fewer boundary edges after 'remove' pruning, all of its edges are
            removed. This parameter is ignored if 'nullify' pruning is used. Defaults to 6.
        :type boundary_edge_minimum: int, optional
        :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
            is set to 1 for all core nodes with boundary edges. Defaults to True.
        :type exact_boundary_outdegree: bool, optional
        :param compute_statistics: Whether to compute variances and confidence intervals. Defaults to True.
        :type compute_statistics: bool, optional
        :param alpha: The confidence level for the confidence intervals. Defaults to 0.95.
        :type alpha: float, optional
        :param permutations: The permutations to test. May contain 'o', 'k1', and 'k2' in any order.
            Defaults to ('o', 'k2').
        :type permutations: list, optional
        :param full_core_permutation: Whether to use the full permutation matrix for each core permutation. Partial
            permutations sample core coefficients, while full permutations sample perturbation scores. Defaults to True.
        :type full_core_permutation: bool, optional
        :param p_iters: The number of permutations to perform. Defaults to 500.
        :type p_iters: int, optional
        :param p_rate: The fraction of boundary coefficients to permute. Defaults to 1.
        :type p_rate: float, optional
        :param sparse: Whether to use sparse computation. Defaults to True.
        :type sparse: bool, optional
        :param seed: The seed for the random number generator. Defaults to None.
        :type seed: int, optional
        :param verbose: Whether to log progress and network statistics. Defaults to True.
        :type verbose: bool, optional
        :param logging_kwargs: The keyword arguments to pass to logging.basicConfig. Defaults to  "stream": sys.stdout,
            "level": logging.INFO, "format": "%(asctime)s %(levelname)s -- %(message)s".
        :type logging_kwargs: dict, optional
        :raises ValueError: If the same node appears in both the core and boundary network.
        :return: The NPA result.
        :rtype: perturbationx.NPAResult
        """
        if verbose and logging_kwargs is not None:
            logging.basicConfig(**logging_kwargs)

        # Copy the graph and set metadata
        prograph = self._graph.to_directed()
        self.initialize_metadata()
        prograph.graph.update(self.metadata)

        return toponpa.toponpa(
            prograph, self.relation_translator, datasets,
            missing_value_pruning_mode=missing_value_pruning_mode,
            opposing_value_pruning_mode=opposing_value_pruning_mode,
            opposing_value_minimum_amplitude=opposing_value_minimum_amplitude,
            boundary_edge_minimum=boundary_edge_minimum,
            exact_boundary_outdegree=exact_boundary_outdegree,
            compute_statistics=compute_statistics, alpha=alpha,
            permutations=permutations, full_core_permutation=full_core_permutation,
            p_iters=p_iters, p_rate=p_rate, sparse=sparse, seed=seed, verbose=verbose
        )

    def to_networkx(self):
        """Return a copy of the network as a NetworkX graph.

        :return: The NetworkX graph.
        :rtype: nx.DiGraph
        """
        return self._graph.copy()

    def to_edge_list(self, edge_type="all", data_cols=DEFAULT_DATA_COLS,):
        """Convert the network to a list of edges.

        :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
        :type edge_type: list | str, optional
        :param data_cols: data_cols: List of data columns to include. Defaults to ["subject", "object",
            "relation", "type"].
        :type data_cols: list, optional
        :return: A list of edges.
        :rtype: list
        """
        return px_io.to_edge_list(self._graph, edge_type, data_cols)

    def to_dsv(self, filepath, edge_type="all", delimiter='\t',
               data_cols=DEFAULT_DATA_COLS, header=None):
        """Write the network to a delimited file.

        :param filepath: The path to write the file to.
        :type filepath: str
        :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
        :type edge_type: list | str, optional
        :param delimiter: The delimiter to use in the DSV file. Defaults to "\t".
        :type delimiter: str, optional
        :param data_cols: List of data columns to include. Columns not in ["subject", "object", "relation", "type"] will
            be ignored. Defaults to ["subject", "object", "relation", "type"].
        :type data_cols: list, optional
        :param header: List of header values to use. Must be given in the same order as data_cols. Defaults to None.
        :type header: list, optional
        :raises ValueError: If the length of the header list does not match the length of the data_cols list.
        """
        px_io.write_dsv(self._graph, filepath, edge_type, delimiter, data_cols, header)

    def to_tsv(self, filepath, edge_type="all",
               data_cols=DEFAULT_DATA_COLS, header=None):
        """Write the network to a tab-separated file.

        :param filepath: The path to write the file to.
        :type filepath: str
        :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
        :type edge_type: list | str, optional
        :param data_cols: List of data columns to include. Columns not in ["subject", "object", "relation", "type"] will
            be ignored. Defaults to ["subject", "object", "relation", "type"].
        :type data_cols: list, optional
        :param header: List of header values to use. Must be given in the same order as data_cols. Defaults to None.
        :type header: list, optional
        :raises ValueError: If the length of the header list does not match the length of the data_cols list.
        """
        self.to_dsv(filepath, edge_type, data_cols=data_cols, header=header)

    def to_csv(self, filepath, edge_type="all",
               data_cols=DEFAULT_DATA_COLS, header=None):
        """Write the network to a comma-separated file.

        :param filepath: The path to write the file to.
        :type filepath: str
        :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
        :type edge_type: list | str, optional
        :param data_cols: List of data columns to include. Columns not in ["subject", "object", "relation", "type"] will
            be ignored. Defaults to ["subject", "object", "relation", "type"].
        :type data_cols: list, optional
        :param header: List of header values to use. Must be given in the same order as data_cols. Defaults to None.
        :type header: list, optional
        :raises ValueError: If the length of the header list does not match the length of the data_cols list.
        """
        self.to_dsv(filepath, edge_type, delimiter=",", data_cols=data_cols, header=header)

    def to_cyjs_json(self, filepath, indent=4):
        """Write the network to a Cytoscape.js JSON file.

        :param filepath: The path to write the file to.
        :type filepath: str
        :param indent: The indentation to use in the JSON file. Defaults to 4.
        :type indent: int, optional
        """
        with open(filepath, 'w') as f:
            json.dump(nx.cytoscape_data(self._graph), f, ensure_ascii=False, indent=indent)



from perturbationx.util import connectivity, neighbors, shortest_paths

from .connectivity import *
from .neighbors import *
from .shortest_paths import *

__all__ = ['connectivity', 'neighbors', 'shortest_paths', 'connect_adjacency_components',
           'get_neighborhood_components', 'get_shortest_path_components']



import warnings

import networkx as nx
import numpy as np
import pandas as pd

from .NPAResult import NPAResult


class NPAResultBuilder:
    """Class for building NPAResult objects. Node attributes can be only passed for nodes in the graph with type "core".
    Each core node should have an attribute "idx" that is a unique integer index for that node. Node attributes should
    be passed as a list of values that is ordered by the node indices.

    :param graph: The graph used to generate the result.
    :type graph: networkx.DiGraph
    :param datasets: The datasets used to generate the result.
    :type datasets: list
    """
    def __init__(self, graph: nx.DiGraph, datasets: list):
        """Construct a new NPAResultBuilder object.
        """
        self._graph = graph
        self._datasets = datasets
        self._nodes = sorted([n for n in graph.nodes if graph.nodes[n]["type"] == "core"],
                             key=lambda n: graph.nodes[n]["idx"])

        self._global_attributes = {}  # One score per dataset
        self._node_attributes = {}  # N scores per dataset, where N is number of nodes
        self._distributions = {}  # K scores per dataset, where K is any positive integer

    @classmethod
    def new_builder(cls, graph: nx.DiGraph, datasets: list):
        """Construct a new NPAResultBuilder object.
        """
        return cls(graph, datasets)

    def set_global_attributes(self, dataset_id: str, attributes: list, values: list):
        """Set global attributes for a dataset. Attributes and values should be ordered in the same way.

        :param dataset_id: Dataset to set attributes for.
        :type dataset_id: str
        :param attributes: List of attribute names.
        :type attributes: list
        :param values: List of attribute values.
        :type values: list
        """
        if dataset_id not in self._datasets:
            warnings.warn("Attributes for unknown dataset %s will be ignored." % dataset_id)
            return

        for attr, val in zip(attributes, values):
            if attr not in self._global_attributes:
                self._global_attributes[attr] = {d: np.nan for d in self._datasets}
            self._global_attributes[attr][dataset_id] = val

    def set_node_attributes(self, dataset_id: str, attributes: list, values: list):
        """Set node attributes for a dataset. Attributes and values should be ordered in the same way.
        Values should be passed as a nested list that is ordered by the node indices.

        :param dataset_id: Dataset to set attributes for.
        :type dataset_id: str
        :param attributes: List of attribute names.
        :type attributes: list
        :param values: List of attribute values.
        :type values: list
        """
        if dataset_id not in self._datasets:
            warnings.warn("Node attributes %s for unknown dataset %s will be ignored."
                          % (str(attributes), dataset_id))
            return

        for attr, val in zip(attributes, values):
            if len(val) != len(self._nodes):
                warnings.warn("Node attribute %s for dataset %s is of incorrect length "
                              "and will be ignored" % (attr, dataset_id))
                continue

            if attr not in self._node_attributes:
                self._node_attributes[attr] = {d: np.full(len(self._nodes), np.nan) for d in self._datasets}
            self._node_attributes[attr][dataset_id] = val

    def set_distribution(self, dataset_id: str, distribution: str, values: list, reference=None):
        """Set a distribution for a dataset.

        :param dataset_id: Dataset to set distribution for.
        :type dataset_id: str
        :param distribution: Name of distribution.
        :type distribution: str
        :param values: List of values.
        :type values: list
        :param reference: Reference value for distribution. Defaults to None.
        :type reference: float, optional
        """
        if dataset_id not in self._datasets:
            warnings.warn("Distribution %s for unknown dataset %s will be ignored."
                          % (distribution, dataset_id))
            return

        if distribution not in self._distributions:
            self._distributions[distribution] = {d: ([], None) for d in self._datasets}
        self._distributions[distribution][dataset_id] = (values, reference)

    def build(self, metadata=None):
        """Construct an NPAResult object.

        :param metadata: Metadata to include in the result. Defaults to None.
        :type metadata: dict, optional
        :return: NPAResult object.
        :rtype: perturbationx.NPAResult
        """
        # Create global info dataframe
        global_info = pd.DataFrame({
            attr: [self._global_attributes[attr][d] for d in self._datasets]
            for attr in self._global_attributes
        })
        if len(self._global_attributes.keys()) > 0:
            global_info = global_info.set_index([self._datasets])

        # Create node info dataframe
        node_indices = pd.MultiIndex.from_product(
            [self._datasets, self._node_attributes.keys()], names=["data", "attr"]
        )
        node_info = pd.DataFrame([
            self._node_attributes[attr][d]
            for d, attr in node_indices
        ], index=node_indices, columns=self._nodes)

        return NPAResult(self._graph, self._datasets, global_info,
                         node_info, self._distributions, metadata)



from perturbationx.toponpa.permutation import permutation, adjacency_permutation

from .permutation import *

__all__ = ['permutation', 'adjacency_permutation', 'permute_adjacency', 'permute_edge_list']



import networkx as nx

__all__ = ["get_shortest_path_components"]


def get_shortest_path_components(graph: nx.DiGraph, endpoints: set, directed=False, length_tolerance=0):
    """Find paths between nodes in a graph and return the nodes and edges in those paths.

    :param graph: The graph to search for paths in.
    :type graph: nx.DiGraph
    :param endpoints: The nodes to find paths between.
    :type endpoints: set
    :param directed: Whether to search for directed or undirected shortest paths.
    :type directed: bool, optional
    :param length_tolerance: The tolerance for the length of the shortest paths. If 0, only the shortest paths
            are returned. If length_tolerance is an integer, it is interpreted as an absolute length. If
            length_tolerance is a float, it is interpreted as a percentage of the length of the shortest path.
            Defaults to 0.
    :type length_tolerance: int | float, optional
    :raises ValueError: If length_tolerance is not a number or is negative.
    :return: The nodes and edges in the paths. They are returned as a pair of lists.
    :rtype: (list, list)
    """

    try:
        if type(length_tolerance) != int and type(length_tolerance) != float:
            length_tolerance = float(length_tolerance)
            if length_tolerance.is_integer():
                length_tolerance = int(length_tolerance)
    except ValueError:
        raise ValueError("Argument length_tolerance must be a number.")

    if length_tolerance < 0.:
        raise ValueError("Argument length_tolerance must be >= 0.")

    dir_graph = graph
    if not directed:
        graph = graph.to_undirected()

    endpoint_pairs = set()
    for n in endpoints:
        for m in endpoints:
            # If undirected graph, we only need to consider one ordering
            if n != m and (directed or (m, n) not in endpoint_pairs):
                endpoint_pairs.add((n, m))

    nodes, edges = set(), set()
    for src, trg in endpoint_pairs:
        try:
            shortest_len = nx.shortest_path_length(graph, source=src, target=trg)

            if type(length_tolerance) == int:
                max_len = shortest_len + length_tolerance
            else:
                max_len = shortest_len * (1. + length_tolerance)

            paths = nx.all_simple_paths(graph, source=src, target=trg, cutoff=max_len)
            for p in paths:
                nodes.update(p)
                for n, m in zip(p, p[1:]):
                    if dir_graph.has_edge(n, m):
                        edges.add((n, m, dir_graph[n][m]["interaction"]))
                    elif dir_graph.has_edge(m, n):
                        edges.add((m, n, dir_graph[m][n]["interaction"]))

        except nx.NetworkXNoPath:
            pass

    return nodes, edges



from perturbationx import io
from perturbationx import resources
from perturbationx import result
from perturbationx import toponpa as topo_npa
from perturbationx import util
from perturbationx import vis

from .CausalNetwork import CausalNetwork
from .io.RelationTranslator import RelationTranslator
from .result.NPAResultBuilder import NPAResultBuilder
from .result.NPAResult import NPAResult
from .vis.NPAResultDisplay import NPAResultDisplay


__all__ = ['io', 'resources', 'result', 'topo_npa', 'util', 'vis', 'CausalNetwork', 'RelationTranslator',
           'NPAResult', 'NPAResultBuilder', 'NPAResultDisplay']



import warnings

import networkx as nx
import numpy as np
from scipy.sparse import issparse, lil_array, csr_array, SparseEfficiencyWarning, sparray

__all__ = ["generate_adjacencies", "generate_boundary_laplacian", "generate_core_laplacians"]


def generate_adjacencies(graph: nx.DiGraph, directed=False, sparse=True):
    """Generate the boundary and core adjacency matrices from a graph.

    :param graph: The graph.
    :type graph: nx.DiGraph
    :param directed: Whether to generate directed adjacency matrices. Defaults to False.
    :type directed: bool, optional
    :param sparse: Whether to generate sparse adjacency matrices. Defaults to True.
    :type sparse: bool, optional
    :return: The boundary and core adjacency matrices.
    :rtype: (np.ndarray, np.ndarray) | (sp.sparray, sp.sparray)
    """
    core_size = sum(1 for n in graph.nodes if graph.nodes[n]["type"] == "core")
    network_size = graph.number_of_nodes()

    rows = [graph.nodes[src]["idx"] for src, trg in graph.edges]
    cols = [graph.nodes[trg]["idx"] for src, trg in graph.edges]
    data = [graph[src][trg]["weight"] for src, trg in graph.edges]

    if sparse:
        adj = lil_array((core_size, network_size))
    else:
        adj = np.zeros((core_size, network_size))

    adj[rows, cols] = data
    adj_c = adj[:core_size, :core_size]
    adj_b = adj[:core_size, core_size:]

    if not directed:
        adj_c += adj_c.transpose()

    return adj_b, adj_c


def generate_boundary_laplacian(adj_b: np.ndarray | sparray, boundary_edge_minimum=6):
    """Generate the boundary Lb Laplacian from a boundary adjacency matrix.

    :param adj_b: The boundary adjacency matrix.
    :type adj_b: np.ndarray | sp.sparray
    :param boundary_edge_minimum: The minimum number of boundary edges a core node must have to be
                                    included in the Lb Laplacian. Nodes with fewer boundary edges
                                    are removed from the Lb Laplacian. Defaults to 6.
    :type boundary_edge_minimum: int, optional
    :raises ValueError: If the adjacency matrix is misshapen or if the boundary edge minimum is negative.
    :return: The boundary Lb Laplacian.
    :rtype: np.ndarray | sp.sparray
    """

    if adj_b.ndim != 2:
        raise ValueError("Argument adj_b is not two-dimensional.")
    if boundary_edge_minimum < 0:
        raise ValueError("Boundary edge minimum must be non-negative.")

    adj_b = adj_b.copy()

    if issparse(adj_b):
        # Clip edges where boundary outdegree is below threshold
        row_nonzero_counts = np.bincount(adj_b.nonzero()[0], minlength=adj_b.shape[0])
        clipping_mask = row_nonzero_counts < boundary_edge_minimum
        clipping_mask *= row_nonzero_counts > 0

        clip_row_idx = [idx for idx, clip_row in enumerate(clipping_mask) if clip_row] + [0]
        for idx in clip_row_idx:
            adj_b[idx, :] = 0

        row_sums = abs(adj_b).sum(axis=1)
        row_sums[row_sums == 0] = 1
        adj_b /= row_sums[:, np.newaxis]
        adj_b = csr_array(adj_b)

    else:
        # Clip edges where boundary outdegree is below threshold
        row_nonzero_counts = np.count_nonzero(adj_b, axis=1)
        adj_b[row_nonzero_counts < boundary_edge_minimum, :] = 0

        row_sums = np.abs(adj_b).sum(axis=1)
        row_sums[row_sums == 0] = 1
        adj_b /= row_sums[:, np.newaxis]

    return adj_b


def generate_core_laplacians(lap_b: np.ndarray | sparray, adj_c: np.ndarray | sparray, exact_boundary_outdegree=True):
    """Generate the core Laplacians from a boundary Laplacian and core adjacency matrix.

    :param lap_b: The boundary Laplacian.
    :type lap_b: np.ndarray | sp.sparray
    :param adj_c: The core adjacency matrix.
    :type adj_c: np.ndarray | sp.sparray
    :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
                                        is set to 1 for all core nodes with boundary edges. Defaults to True.
    :type exact_boundary_outdegree: bool, optional
    :return: The core Laplacians.
    :rtype: (np.ndarray, np.ndarray) | (sp.sparray, sp.sparray)
    """
    core_degrees = np.abs(adj_c).sum(axis=1)
    boundary_outdegrees = np.abs(lap_b).sum(axis=1)

    if not exact_boundary_outdegree:
        # Fix the boundary outdegrees to 1 if they are non-zero
        boundary_outdegrees[boundary_outdegrees != 0] = 1

    if issparse(adj_c):
        lap_c = - adj_c.copy()
        lap_q = adj_c.copy()

        # lap_c is always converted to csr when inverting the sign (-)
        # this causes a SparseEfficiencyWarning when setting the diagonal
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=SparseEfficiencyWarning)
            lap_c.setdiag(core_degrees + boundary_outdegrees)
            lap_q.setdiag(core_degrees)
            lap_q = lap_q.tocsr()

    else:
        lap_c = np.diag(core_degrees + boundary_outdegrees) - adj_c
        lap_q = np.diag(core_degrees) + adj_c

    return lap_c, lap_q



import numpy as np
import numpy.linalg as la
from scipy.sparse import issparse, lil_array, sparray


def core_covariance_matrix(lap_b: np.ndarray | sparray, lap_c: np.ndarray | sparray, stderr: np.ndarray):
    """Compute the covariance matrix of the core node coefficients.

    :param lap_b: The Lb boundary Laplacian.
    :type lap_b: np.ndarray | sp.sparray
    :param lap_c: The Lc core Laplacian.
    :type lap_c: np.ndarray | sp.sparray
    :param stderr: The standard error of the boundary coefficients.
    :type stderr: np.ndarray
    :return: The covariance matrix of the core node coefficients.
    :rtype: np.ndarray
    """
    if issparse(lap_c):
        lap_c = lap_c.todense()
    lap_c_inv = la.inv(lap_c)

    err_diag = lil_array((stderr.shape[0], stderr.shape[0]))
    err_diag.setdiag(stderr ** 2)
    tmp = lap_b @ err_diag @ lap_b.T
    res = lap_c_inv @ tmp @ lap_c_inv.T
    return res


def perturbation_variance(lap_q: np.ndarray | sparray, core_coefficients: np.ndarray,
                          core_covariance: np.ndarray, core_edge_count: int):
    """Compute the variance of the perturbation score.

    :param lap_q: The Q core Laplacian.
    :type lap_q: np.ndarray | sp.sparray
    :param core_coefficients: The core node coefficients.
    :type core_coefficients: np.ndarray
    :param core_covariance: The covariance matrix of the core node coefficients.
    :type core_covariance: np.ndarray
    :param core_edge_count: The number of edges in the core network.
    :type core_edge_count: int
    :return: The variance of the perturbation score.
    :rtype: float
    """
    temp = lap_q @ core_covariance
    unscaled_variance = 2 * np.trace(temp @ temp) + \
                        4 * core_coefficients @ temp @ lap_q @ core_coefficients
    return unscaled_variance / core_edge_count ** 2




defaultMap = {
    "1": 1.,
    "increases": 1.,
    "->": 1.,
    "directlyIncreases": 1.,
    "=>": 1.,
    "-1": -1.,
    "decreases": -1.,
    "-|": -1.,
    "directlyDecreases": -1.,
    "=|": -1.
}


class RelationTranslator:
    """Class for translating relations to numeric values. By default, relations "1", "increases", "-\>",
    "directlyIncreases", and "=\>" are mapped to 1.0, while relations "-1", "decreases", "-\|", "directlyDecreases",
    and "=\|" are mapped to -1.0. Relations that cannot be mapped to a numeric value will be parsed as 0.0.

    :param mappings: Dictionary of additional relation to numeric value mappings. It extends and overrides the default
        mappings.
    :type mappings: dict, optional
    :param allow_numeric: If True, relations will be parsed as numeric values if they cannot be found in the mappings
        dictionary. Defaults to True.
    :type allow_numeric: bool, optional
    """

    def __init__(self, mappings=None, allow_numeric=True):
        """Construct a new RelationTranslator.
        """
        self._map = defaultMap.copy()
        if mappings is not None:
            for k, v in mappings.items():
                self._map[k] = float(v)
        self._allow_numeric = allow_numeric

    def add_mapping(self, relation, maps_to):
        """Add a new mapping from a relation to a numeric value.

        :param relation: The relation to map.
        :type relation: str
        :param maps_to: The numeric value to map to.
        :type maps_to: float
        """
        self._map[relation] = float(maps_to)

    def remove_mapping(self, relation):
        """Remove a mapping from a relation to a numeric value.

        :param relation: The relation to remove the mapping for.
        :type relation: str
        """
        del self._map[relation]

    def copy(self):
        """Create a copy of this RelationTranslator.

        :return: A copy of this RelationTranslator.
        :rtype: RelationTranslator
        """
        return RelationTranslator(self._map.copy(), self._allow_numeric)

    def translate(self, relation):
        """Translate a relation to a numeric value.

        :param relation: The relation to translate.
        :type relation: str
        :return: The numeric value that the relation maps to.
        :rtype: float
        """

        if relation in self._map:
            res = self._map[relation]
        elif self._allow_numeric:
            try:
                res = float(relation)
            except ValueError:
                res = 0.
        else:
            res = 0.

        return res



from perturbationx.toponpa.statistics import permutation_tests, statistics, variance

from .statistics import *

__all__ = ['permutation_tests', 'statistics', 'variance', 'compute_variances',
           'confidence_interval', 'test_permutations', 'p_value']



import numpy as np
from scipy.sparse import issparse, lil_array, sparray

from perturbationx.util import connect_adjacency_components


def adjacency_permutation_k1(adj: np.ndarray | sparray, iterations=500, permutation_rate=1.,
                             ensure_connectedness=True, seed=None):
    """Permute the edges of an adjacency matrix using the "K1" method. This method permutes edges randomly.

    :param adj: The adjacency matrix to permute.
    :type adj: np.ndarray | sp.sparray
    :param iterations: The number of permutations to generate. Defaults to 500.
    :type iterations: int, optional
    :param permutation_rate: The fraction of edges to permute. Defaults to 1.
    :type permutation_rate: float, optional
    :param ensure_connectedness: Whether to ensure that the permuted adjacency matrix is connected. Defaults to True.
    :type ensure_connectedness: bool, optional
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    :raises ValueError: If the adjacency matrix is not square.
    :return: A list of permuted adjacency matrices.
    :rtype: list
    """
    if adj.ndim != 2 or adj.shape[0] != adj.shape[1]:
        raise ValueError("Argument 'adj' {} is not a square matrix.".format(str(adj.shape)))

    rng = np.random.default_rng(seed)
    tril_idx = np.tril_indices(adj.shape[0], -1)
    adj_tril = adj[tril_idx]
    if issparse(adj_tril):
        adj_tril = adj_tril.todense().flatten()

    edge_idx = np.nonzero(adj_tril)[0]
    edge_weights = adj_tril[edge_idx]
    fixed_edges = np.floor(edge_idx.size * (1. - permutation_rate)).astype(int)

    permuted = []
    for _ in range(iterations):
        fixed_idx = rng.choice(edge_idx, size=fixed_edges, replace=False)
        remaining_idx = [idx for idx in edge_idx if idx not in fixed_idx]
        fixed_weights = adj_tril[fixed_idx]
        remaining_weights = adj_tril[remaining_idx]

        permuted_idx = np.arange(adj_tril.size)
        permuted_idx = np.delete(permuted_idx, fixed_idx)
        permuted_idx = rng.choice(permuted_idx, size=remaining_weights.size, replace=False)

        combined_idx = np.concatenate((fixed_idx, permuted_idx))
        combined_weights = np.concatenate((fixed_weights, remaining_weights))

        if issparse(adj):
            random_adj = lil_array(adj.shape)
        else:
            random_adj = np.zeros(adj.shape)

        src_indices = tril_idx[0][combined_idx]
        trg_indices = tril_idx[1][combined_idx]
        random_adj[src_indices, trg_indices] = combined_weights
        random_adj[trg_indices, src_indices] += combined_weights

        if ensure_connectedness:
            connect_adjacency_components(random_adj, weights=edge_weights, seed=seed)
        permuted.append(random_adj)

    return permuted


def adjacency_permutation_k2(adj: np.ndarray | sparray, iterations=500, permutation_rate=1.,
                             ensure_connectedness=True, seed=None):
    """Permute the edges of an adjacency matrix using the "K2" method. This method permutes edges by preserving
    the degree of each node as much as possible.

    :param adj: The adjacency matrix to permute.
    :type adj: np.ndarray | sp.sparray
    :param iterations: The number of permutations to generate. Defaults to 500.
    :type iterations: int, optional
    :param permutation_rate: The fraction of edges to permute. Defaults to 1.
    :type permutation_rate: float, optional
    :param ensure_connectedness: Whether to ensure that the permuted adjacency matrix is connected. Defaults to True.
    :type ensure_connectedness: bool, optional
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    :raises ValueError: If the adjacency matrix is not square.
    :return: A list of permuted adjacency matrices.
    :rtype: list
    """
    if adj.ndim != 2 or adj.shape[0] != adj.shape[1]:
        raise ValueError("Argument 'adj' {} is not a square matrix.".format(str(adj.shape)))

    rng = np.random.default_rng(seed)
    edge_stubs = np.array([[idx, idy] for idx, idy in zip(*np.nonzero(adj)) if idx < idy])

    total_edges = edge_stubs.shape[0]
    permuted_edges = np.ceil(total_edges * permutation_rate).astype(int)
    edge_weights = np.array([adj[idx, idy] for idx, idy in edge_stubs])

    permuted = []
    for _ in range(iterations):
        permuted_idx = rng.choice(total_edges, size=permuted_edges, replace=False)
        permuted_stubs = edge_stubs.copy()
        permuted_stubs[permuted_idx, 0] = rng.permutation(permuted_stubs[permuted_idx, 0])
        permuted_stubs[permuted_idx, 1] = rng.permutation(permuted_stubs[permuted_idx, 1])

        permuted_edge_weights = edge_weights.copy()
        permuted_edge_weights[permuted_idx] = rng.permutation(permuted_edge_weights[permuted_idx])

        if issparse(adj):
            random_adj = lil_array(adj.shape)
        else:
            random_adj = np.zeros(adj.shape)

        src_indices, trg_indices = zip(*permuted_stubs)
        mask = src_indices != trg_indices

        src_indices = np.array(src_indices)[mask]
        trg_indices = np.array(trg_indices)[mask]
        permuted_edge_weights = permuted_edge_weights[mask]

        random_adj[src_indices, trg_indices] = permuted_edge_weights
        random_adj[trg_indices, src_indices] += permuted_edge_weights

        if ensure_connectedness:
            connect_adjacency_components(random_adj, weights=edge_weights, seed=seed)

        permuted.append(random_adj)

    return permuted



import sys
import os
import logging

__all__ = [
    "DEFAULT_DATA_COLS", "DEFAULT_LOGGING_KWARGS", "STYLE_XML", "DEFAULT_STYLE", "DEFAULT_NODE_COLOR",
    "DEFAULT_GRADIENT", "DEFAULT_EDGE_WIDTH", "DEFAULT_NODE_BORDER_WIDTH", "get_style_xml_path",
    "get_osmotic_stress_core_path", "get_osmotic_stress_boundary_path", "get_osmotic_stress_dataset_paths"
]


# io
DEFAULT_DATA_COLS = ("subject", "object", "relation", "type")
DEFAULT_LOGGING_KWARGS = {
    "stream": sys.stdout,
    "level": logging.INFO,
    "format": "%(asctime)s %(levelname)s -- %(message)s",
}


# visualisation
STYLE_XML = "default-style.xml"
DEFAULT_STYLE = "perturbationx-default"
DEFAULT_NODE_COLOR = "#FEE391"
DEFAULT_GRADIENT = ("#2B80EF", "#EF3B2C")
DEFAULT_EDGE_WIDTH = 2.0
DEFAULT_NODE_BORDER_WIDTH = 4.0


def get_style_xml_path():
    """Get the path to the default style xml file.
    :return: The path to the default style xml file.
    :rtype: str
    """
    module_path = os.path.abspath(__file__)
    resource_path = os.path.join(os.path.dirname(module_path), STYLE_XML)
    return resource_path


def get_osmotic_stress_core_path():
    """Get the path to the osmotic stress core network file. The network is a subset of the human osmotic stress network
    from the Causal Biological Networks database (https://www.causalbionet.com/, network ID osmotic_stress_2.0_hs).
    All non-causal edges were removed and the largest connected component was then selected.

    :return: The path to the osmotic stress core network file.
    :rtype: str
    """
    module_path = os.path.abspath(__file__)
    resource_path = os.path.join(os.path.dirname(module_path), "osmotic_stress_core.tsv")
    return resource_path


def get_osmotic_stress_boundary_path():
    """Get the path to the osmotic stress boundary network file. The network was generated using the Barabási-Albert
    model and is not biological in nature. It contains 100 outer boundary nodes and 500 boundary edges.

    :return: The path to the osmotic stress boundary network file.
    :rtype: str
    """
    module_path = os.path.abspath(__file__)
    resource_path = os.path.join(os.path.dirname(module_path), "osmotic_stress_boundary.tsv")
    return resource_path


def get_osmotic_stress_dataset_paths():
    """Get the paths to the osmotic stress dataset files. The datasets were generated using a genetic algorithm and
    are not biological in nature.

    :return: The paths to the osmotic stress dataset files.
    :rtype: list
    """
    module_path = os.path.abspath(__file__)
    resource_paths = [
        os.path.join(os.path.dirname(module_path), "osmotic_stress_dataset_" + idx + ".tsv")
        for idx in ["1", "2", "3"]
    ]
    return resource_paths



import csv
import warnings

import networkx as nx
import pandas

from perturbationx.resources.resources import DEFAULT_DATA_COLS


__all__ = ["read_dsv", "parse_pandas", "validate_nx_graph", "to_edge_list", "write_dsv"]


def read_dsv(filepath: str, default_edge_type="infer", delimiter='\t', header_cols=DEFAULT_DATA_COLS):
    """Read a DSV file and return a list of edges.

    :param filepath: Path to the DSV file.
    :type filepath: str
    :param default_edge_type: The default edge type to use if none is specified in the file. Defaults to "infer".
    :type default_edge_type: str, optional
    :param delimiter: The delimiter used in the DSV file. Defaults to "\t".
    :type delimiter: str, optional
    :param header_cols: The column names of the columns in the DSV file. Defaults to ["subject", "object", "relation",
        "type"].
    :type header_cols: list, optional
    :return: A list of edges.
    :rtype: list
    """
    if default_edge_type is None:
        default_edge_type = "infer"
    edge_dict = dict()

    with open(filepath, newline='') as input_file:
        dsv_file = csv.reader(input_file, delimiter=delimiter)
        header = [h.lower() for h in next(dsv_file)]

        # check if columns for subject, object and relation are present
        if header_cols is not None and set(header_cols[0:2]).issubset(header):
            # determine location of columns based on header
            src_idx, trg_idx, rel_idx, typ_idx = header.index(header_cols[0]), \
                header.index(header_cols[1]), header.index(header_cols[2]), None
            if len(header_cols) > 3 and header_cols[3] in header:
                typ_idx = header.index(header_cols[3])

        else:  # set to default values
            src_idx, trg_idx, rel_idx, typ_idx = 0, 1, 2, None
            if len(header) > 3:
                typ_idx = 3
            input_file.seek(0)

        # parse edges
        for line in dsv_file:
            if len(line) == 0:
                continue

            src, rel, trg = line[src_idx], line[rel_idx], line[trg_idx]
            typ = line[typ_idx] if typ_idx is not None else default_edge_type

            if (src, trg) in edge_dict:
                warnings.warn("Duplicate edge (%s, %s) found in file %s. "
                              "Only the first occurrence will be used." % (src, trg, filepath))
                continue

            edge_dict[(src, trg)] = (rel, typ)

    edge_list = [(src, trg, {"relation": rel, "type": typ})
                 for (src, trg), (rel, typ) in edge_dict.items()]
    return edge_list


def parse_pandas(df: pandas.DataFrame, default_edge_type="infer", header_cols=DEFAULT_DATA_COLS):
    """Parse a pandas DataFrame and return a list of edges.

    :param df: The DataFrame to parse.
    :type df: pandas.DataFrame
    :param default_edge_type: The default edge type to use if none is specified in the DataFrame. Defaults to "infer".
    :type default_edge_type: str, optional
    :param header_cols: The column names of the columns in the DataFrame. Defaults to ["subject", "object", "relation",
        "type"].
    :type header_cols: list, optional
    :return: A list of edges.
    :rtype: list
    """
    if header_cols is not None and set(header_cols[0:2]).issubset(df.columns):
        src_col, trg_col, rel_col, typ_col = header_cols[0], header_cols[1], header_cols[2], None
        if len(header_cols) > 3 and header_cols[3] in df.columns:
            typ_col = header_cols[3]
    else:
        src_col, trg_col, rel_col, typ_col = df.columns[0], df.columns[1], df.columns[2], None
        if len(df.columns) > 3:
            typ_col = df.columns[3]

    df_copy = df.copy()
    if typ_col is None:
        df_copy["type"] = default_edge_type
        typ_col = "type"
    else:
        df_copy[typ_col][df_copy[typ_col].isna()] = default_edge_type

    edge_list = df_copy[[src_col, trg_col, rel_col, typ_col]].to_records(index=False)
    edge_list = [(src, trg, {"relation": rel, "type": typ}) for src, trg, rel, typ in edge_list]
    return edge_list


def validate_nx_graph(graph: nx.DiGraph, allowed_edge_types: list):
    """Validate a networkx.DiGraph object for use with toponpa.

    :param graph: The graph to validate.
    :type graph: nx.DiGraph
    :param allowed_edge_types: List of allowed edge types. Edges with types not in this list will be replaced with
        "infer".
    :type allowed_edge_types: list
    :raises TypeError: If the graph is not a networkx.DiGraph.
    """
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("Argument graph is not a networkx.Digraph.")

    to_remove = []
    for src, trg, data in graph.edges.data():
        if "relation" not in data:
            warnings.warn("Edge between %s and %s does not have a "
                          "relation specified and will be ignored." % (src, trg))
            to_remove.append((src, trg))
            continue

        if "type" not in data:
            graph[src][trg]["type"] = "infer"
        elif data["type"] not in allowed_edge_types:
            warnings.warn("Unknown type %s of edge %s will be replaced "
                          "with \"infer\"." % (data["type"], str((src, trg))))
            graph[src][trg]["type"] = "infer"

    for src, trg in to_remove:
        graph.remove_edge(src, trg)
        if graph.degree[src] == 0:
            graph.remove_node(src)
        if graph.degree[trg] == 0:
            graph.remove_node(trg)


def to_edge_list(graph: nx.DiGraph, edge_type="all", data_cols=DEFAULT_DATA_COLS):
    """Convert a networkx.DiGraph to a list of edges.

    :param graph: The graph to convert.
    :type graph: nx.DiGraph
    :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
    :type edge_type: list | str, optional
    :param data_cols: List of data columns to include. Defaults to ["subject", "object", "relation", "type"].
    :type data_cols: list, optional
    :return: A list of edges.
    :rtype: list
    """
    edge_list = nx.to_edgelist(graph)

    # ensure that data_cols contains valid values,
    d_cols = [c for c in data_cols if c in DEFAULT_DATA_COLS]
    if len(d_cols) < len(data_cols):
        warnings.warn("Data columns %s are invalid and will be ignored. "
                      "Please use 'subject', 'object', 'relation' and 'type'."
                      % str(tuple(c for c in data_cols if c not in d_cols)))

    # compute data index to ease line construction
    d_idx = [DEFAULT_DATA_COLS.index(c) for c in d_cols]

    # filter edge list based on edge type
    if edge_type != "all":
        edge_list = [e for e in edge_list if e[2]["type"] == edge_type]

    # format edge list
    edge_list = [(e[0], e[1], e[2]["relation"], e[2]["type"]) for e in edge_list]
    edge_list = [tuple([e[i] for i in d_idx]) for e in edge_list]

    return edge_list


def write_dsv(graph: nx.DiGraph, filepath: str, edge_type="all", delimiter='\t',
              data_cols=DEFAULT_DATA_COLS, header=None):
    """Write a networkx.DiGraph to a delimited file.

    :param graph: The graph to write.
    :type graph: nx.DiGraph
    :param filepath: The path to write the file to.
    :type filepath: str
    :param edge_type: List of edge types to include. If "all", all edges will be included. Defaults to "all".
    :type edge_type: list | str, optional
    :param delimiter: The delimiter to use in the DSV file. Defaults to "\t".
    :type delimiter: str, optional
    :param data_cols: List of data columns to include. Columns not in ["subject", "object", "relation", "type"] will
        be ignored. Defaults to ["subject", "object", "relation", "type"].
    :type data_cols: list, optional
    :param header: List of header values to use. Must be given in the same order as data_cols. Defaults to None.
    :type header: list, optional
    :raises ValueError: If the length of the header list does not match the length of the data_cols list.
    """
    if header is not None and len(header) != len(data_cols):
        raise ValueError("Please pass header values for all columns.")

    # construct edge list
    edge_list = to_edge_list(graph, edge_type=edge_type, data_cols=data_cols)

    # shorten header by skipping ignored columns
    d_cols = [c for c in data_cols if c in DEFAULT_DATA_COLS]
    if header is not None:
        header = [h for h, c in zip(header, data_cols) if c in d_cols]

    with open(filepath, 'w', newline='') as output_file:
        dsv_file = csv.writer(output_file, delimiter=delimiter)
        if header is not None:
            dsv_file.writerow(header)

        dsv_file.writerows(edge_list)



import numpy as np

__all__ = ["connect_adjacency_components"]


def get_root(node, parents: dict):
    """Get the root of a node in a tree represented by a dictionary of parent-child relationships.

    :param node: The node whose root to find.
    :param parents: A dictionary of parent-child relationships.
    :type parents: dict
    :return: The root of the node.
    """
    if parents[node] == node:
        return node

    parents[node] = get_root(parents[node], parents)
    return parents[node]


def find_connected_components(adj: np.ndarray):
    """Find the connected components of a graph represented by an adjacency matrix.

    :param adj: The adjacency matrix of the graph.
    :type adj: np.ndarray
    :raises ValueError: If the adjacency matrix is not square.
    :return: A dictionary of connected components, where the keys are the roots of the components and the values are
        the nodes in the components.
    :rtype: dict
    """
    if adj.ndim != 2 or adj.shape[0] != adj.shape[1]:
        raise ValueError("Argument 'adj' {} is not a square matrix.".format(str(adj.shape)))

    num_nodes = adj.shape[0]
    parents = {idx: idx for idx in range(num_nodes)}
    for idx, idy in zip(*np.nonzero(adj)):
        if idx < idy:
            root_x = get_root(idx, parents)
            root_y = get_root(idy, parents)

            if root_x != root_y:
                parents[root_y] = root_x

    components = {}
    for idx in range(num_nodes):
        root = get_root(idx, parents)
        if root not in components:
            components[root] = set()
        components[root].add(idx)

    return components


def connect_adjacency_components(adj: np.ndarray, nodes=None, weights=(1.,), seed=None):
    """Connect the components of a graph represented by an adjacency matrix.

    :param adj: The adjacency matrix of the graph.
    :type adj: np.ndarray
    :param nodes: The nodes to connect. If specified, edges will only be added between these nodes (when possible).
    :type nodes: list, optional
    :param weights: The weights to assign to the edges. Each edge will be assigned a random weight from this list.
    :type weights: list, optional
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    :raises ValueError: If the adjacency matrix is not square.
    """
    if adj.ndim != 2:
        raise ValueError("Argument 'adj' {} is not a square matrix.".format(str(adj.shape)))

    components = find_connected_components(adj)

    rng = np.random.default_rng(seed)
    component_roots = rng.permutation(list(components.keys()))
    weights = list(rng.choice(weights, size=len(components) - 1, replace=True))

    for root_x, root_y in zip(component_roots[:-1], component_roots[1:]):
        if nodes is not None:
            component_x = components[root_x].intersection(nodes)
            if len(component_x) == 0:
                component_x = components[root_x]

            component_y = components[root_y].intersection(nodes)
            if len(component_y) == 0:
                component_y = components[root_y]
        else:
            component_x = components[root_x]
            component_y = components[root_y]

        node_x = rng.choice(list(component_x))
        node_y = rng.choice(list(component_y))

        weight = weights.pop()
        adj[node_x, node_y] = weight
        adj[node_y, node_x] = weight

        components[root_y].update(components[root_x])
        del components[root_x]



import warnings

import numpy as np
from scipy.sparse import sparray

from .adjacency_permutation import adjacency_permutation_k1, adjacency_permutation_k2

__all__ = ["permute_adjacency", "permute_edge_list"]


def permute_adjacency(adj: np.ndarray | sparray, permutations=('k2',), iterations=500, permutation_rate=1., seed=None):
    """Permute an adjacency matrix.

    :param adj: The adjacency matrix to permute.
    :type adj: np.ndarray | sp.sparray
    :param permutations: The permutations to apply. May contain 'k1' and 'k2' in any order. Defaults to ('k2',).
    :type permutations: list, optional
    :param iterations: The number of permutations to generate. Defaults to 500.
    :type iterations: int, optional
    :param permutation_rate: The fraction of edges to permute. Defaults to 1.
    :type permutation_rate: float, optional
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    :raises ValueError: If the adjacency matrix is not square.
    :return: A dictionary of lists with permuted adjacency matrices, keyed by the permutation name.
    :rtype: dict
    """
    if type(permutations) is str:
        permutations = [permutations]

    adj_perms = dict()
    for p in set(permutations):
        match p.lower():
            case 'k1':
                adj_perms[p] = adjacency_permutation_k1(
                    adj, iterations=iterations, permutation_rate=permutation_rate,
                    ensure_connectedness=True, seed=seed
                )
            case 'k2':
                adj_perms[p] = adjacency_permutation_k2(
                    adj, iterations=iterations, permutation_rate=permutation_rate,
                    ensure_connectedness=True, seed=seed
                )
            case 'o':
                # Permutation 'o' is not applied to the laplacian.
                continue
            case _:
                warnings.warn("Permutation %s is unknown and will be skipped." % p)

    return adj_perms


def permute_edge_list(edge_list: np.ndarray, node_list=None, iterations=500,
                      method='k1', permutation_rate=1., seed=None):
    """Permute an edge list.

    :param edge_list: The edge list to permute. Must be a 2D array with shape (n_edges, 4). The first two columns
        contain the source and target nodes, the third column contains the edge type, and the fourth column contains
        the confidence weight. Confidence weights are optional.
    :type edge_list: np.ndarray
    :param node_list: The list of nodes to use in the permutation. Only edges that connect nodes in this list
        are permuted. If None, the list is inferred from the edge list.
    :type node_list: list, optional
    :param iterations: The number of permutations to generate. Defaults to 500.
    :type iterations: int, optional
    :param method: The permutation method to use. Defaults to 'k1'. May be 'k1' or 'k2'.
    :type method: str, optional
    :param permutation_rate: The fraction of edges to permute. Defaults to 1. If 'confidence', the confidence weights
        are used to determine the number of edges to permute. For each edge, a random number is drawn from a uniform
        distribution between 0 and 1. If the confidence weight is larger than this number, the edge is permuted.
    :type permutation_rate: float | str, optional
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    :raises ValueError: If the permutation method is unknown.
    :return: A list of permutations. Each permutation is a list of tuples with the source node, target node, and edge
        type. If the edge type is None, the edge is removed.
    """
    if node_list is None:
        node_list = np.unique(edge_list[:, :2])

    if permutation_rate != "confidence":
        permutation_rate = float(permutation_rate)

    permuted_edge_count = np.ceil(edge_list.shape[0] * permutation_rate).astype(int) \
        if type(permutation_rate) is float else None
    confidence_weights = edge_list[:, 3].astype(float)
    rng = np.random.default_rng(seed)

    permutations = []
    for _ in range(iterations):
        permutation = []
        permutations.append(permutation)

        if permutation_rate == "confidence":
            permuted_edge_idx = np.where(rng.uniform(size=edge_list.shape[0]) > confidence_weights)[0]
            permuted_edge_count = permuted_edge_idx.shape[0]
        else:
            permuted_edge_idx = rng.choice(edge_list.shape[0], size=permuted_edge_count, replace=False, axis=0)

        permuted_edges = edge_list[permuted_edge_idx, :].copy()
        fixed_edges = np.delete(edge_list, permuted_edge_idx, axis=0)

        match method.lower():
            case 'k1':
                permuted_edges[:, 0] = rng.choice(node_list, size=permuted_edge_count, replace=True)
                permuted_edges[:, 1] = rng.choice(node_list, size=permuted_edge_count, replace=True)
            case 'k2':
                permuted_edges[:, 0] = rng.permutation(permuted_edges[:, 0])
                permuted_edges[:, 1] = rng.permutation(permuted_edges[:, 1])
            case _:
                raise ValueError("Unknown permutation %s." % permutation)

        permuted_edges[:, 2] = rng.permutation(permuted_edges[:, 2])
        for i in range(permuted_edges.shape[0]):
            src, trg, rel = permuted_edges[i, :3]
            if src != trg and not any(e[0] == src and e[1] == trg for e in permutation):
                permutation.append((src, trg, rel))

        permuted_edges = np.concatenate([permuted_edges, fixed_edges], axis=0)
        for i in range(edge_list.shape[0]):
            src, trg = edge_list[i, :2]
            if not any(permuted_edges[j, 0] == src and permuted_edges[j, 1] == trg
                       for j in range(permuted_edges.shape[0])):
                permutation.append((src, trg, None))

        if len(permutation) == 0:
            warnings.warn("Edge list permutation '%s' produced empty modification list." % method)

    return permutations



import sys
import io
from typing import Optional
from contextlib import redirect_stderr

import pandas as pd
import py4cytoscape as p4c
import networkx as nx
from py4cytoscape.exceptions import CyError
from py4cytoscape.py4cytoscape_utils import DEFAULT_BASE_URL
from py4cytoscape.tables import load_table_data

from perturbationx.resources import *


__all__ = ["edge_to_p4c_format", "init_cytoscape", "load_network_data", "set_boundary_display",
           "color_nodes_by_column", "highlight_subgraph", "isolate_subgraph", "extract_subgraph", "clear_bypass"]


def edge_to_p4c_format(src: str, trg: str, interaction: str, escape_commas=False):
    """Convert an edge to the format used by py4cytoscape.

    :param src: Source node name.
    :type src: str
    :param trg: Target node name.
    :type trg: str
    :param interaction: Interaction type.
    :type interaction: str
    :param escape_commas: Whether to escape commas. Defaults to False.
    :type escape_commas: bool, optional
    :return: Edge name.
    :rtype: str
    """
    # Edge name format used by p4c is 'source (interaction) target'
    edge_formatted = "%s (%s) %s" % (src, interaction, trg)
    if escape_commas:
        edge_formatted = edge_formatted.replace(",", r"\,")
    return edge_formatted


def init_cytoscape(graph: nx.Graph, title: str, collection: str, init_boundary: Optional[bool] = False,
                   network_suid=None, cytoscape_url=DEFAULT_BASE_URL):
    """Initialize a Cytoscape network from a NetworkX graph.

    :param graph: NetworkX graph.
    :type graph: nx.Graph
    :param title: Network title.
    :type title: str
    :param collection: Network collection.
    :type collection: str
    :param init_boundary: Whether to initialize boundary nodes. Defaults to False.
    :type init_boundary: bool, optional
    :param network_suid: Network SUID to update. Defaults to None. If None or invalid, a new network is created.
    :type network_suid: int, optional
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    :return: Network SUID.
    :rtype: int
    """
    if network_suid is not None:
        try:
            p4c.set_current_network(network_suid, base_url=cytoscape_url)
        except CyError:
            # Network does not exist anymore
            network_suid = None

    if network_suid is None:
        # If boundary nodes should not be included, remove them from the graph
        if not init_boundary:
            graph = graph.copy()
            boundary_nodes = [n for n in graph.nodes if graph.nodes[n]["type"] == "boundary"]
            graph.remove_nodes_from(boundary_nodes)
        network_suid = p4c.networks.create_network_from_networkx(
            graph, base_url=cytoscape_url, title=title, collection=collection
        )

    # Import default style if not already present
    if DEFAULT_STYLE not in p4c.styles.get_visual_style_names(base_url=cytoscape_url):
        p4c.styles.import_visual_styles(filename=get_style_xml_path(), base_url=cytoscape_url)

    return network_suid


def load_network_data(dataframe: pd.DataFrame, table: str, network_suid: int, cytoscape_url=DEFAULT_BASE_URL):
    """Load a pandas DataFrame into a Cytoscape table.

    :param dataframe: The DataFrame to load.
    :type dataframe: pd.DataFrame
    :param table: The table to update.
    :type table: str
    :param network_suid: The network to update.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """
    dataframe = dataframe.copy()
    # Rename MultiIndex columns
    dataframe.columns = [col if isinstance(col, str) else ' '.join(col)
                         for col in dataframe.columns]
    for c in dataframe:
        column = dataframe[[c]].dropna().reset_index()
        if column.shape[0] > 0:
            load_table_data(
                column, data_key_column="index", table=table,
                network=network_suid, base_url=cytoscape_url
            )


def set_boundary_display(graph: nx.Graph, show_boundary: bool, network_suid, cytoscape_url=DEFAULT_BASE_URL):
    """Set the display of boundary nodes.

    :param graph: The graph to load boundary nodes from.
    :type graph: nx.Graph
    :param show_boundary: Whether to show boundary nodes. If boundary nodes were not loaded during initialization,
                            they can be loaded in here. If boundary nodes have already been loaded, they are hidden.
    :type show_boundary: bool
    :param network_suid: The network to update.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """
    node_table = p4c.tables.get_table_columns(table="node", columns=["name", "type"],
                                              network=network_suid, base_url=cytoscape_url)
    boundary_nodes = node_table.loc[node_table["type"] == "boundary"]["name"].tolist()

    if len(boundary_nodes) > 0 and show_boundary is False:
        p4c.style_bypasses.hide_nodes(boundary_nodes, network=network_suid, base_url=cytoscape_url)
    elif len(boundary_nodes) > 0 and show_boundary is True:
        # The unhide function deletes the bypass, which throws an error if the bypass does not exist
        p4c.style_bypasses.set_node_property_bypass(
            boundary_nodes, new_values="true", visual_property="NODE_VISIBLE",
            network=network_suid, base_url=cytoscape_url
        )
    elif show_boundary is True:
        # No boundary nodes in Cytoscape :( We need to load them in
        boundary_nodes = [n for n in graph.nodes if graph.nodes[n]["type"] == "boundary"]
        p4c.networks.add_cy_nodes(boundary_nodes, network=network_suid, base_url=cytoscape_url)

        # Find boundary edges and all interaction types
        boundary_edges = [(src, trg) for src, trg in graph.edges
                          if graph[src][trg]["type"] == "boundary"]
        interactions = {graph[src][trg]["interaction"] for src, trg in boundary_edges}

        # Load edges per interaction
        for interaction in interactions:
            boundary_edge_subset = [(src, trg) for src, trg in boundary_edges
                                    if graph[src][trg]["interaction"] == interaction]
            if len(boundary_edge_subset) > 0:
                p4c.networks.add_cy_edges(boundary_edge_subset, edge_type=interaction, directed=True,
                                          network=network_suid, base_url=cytoscape_url)

        # Load node data
        node_data = pd.DataFrame.from_dict({
            n: graph.nodes[n] for n in boundary_nodes
        }, orient="index")
        load_network_data(node_data, "node", network_suid, cytoscape_url)

        # Load edge data
        edge_data = pd.DataFrame.from_dict({
            edge_to_p4c_format(src, trg, graph[src][trg]["interaction"]): graph[src][trg]
            for src, trg in boundary_edges
        }, orient="index")
        load_network_data(edge_data, "edge", network_suid, cytoscape_url)


def color_nodes_by_column(data_column: str, network_suid: int, gradient=DEFAULT_GRADIENT,
                          default_color=DEFAULT_NODE_COLOR, style=DEFAULT_STYLE, cytoscape_url=DEFAULT_BASE_URL):
    """Color nodes by a column in the node table.

    :param data_column: The column to color by.
    :type data_column: str
    :param network_suid: The network to update.
    :type network_suid: int
    :param gradient: The gradient to use. Defaults to DEFAULT_GRADIENT ("#2B80EF", "#EF3B2C").
    :type gradient: (str, str), optional
    :param default_color: The default node color to use. Defaults to DEFAULT_NODE_COLOR ("#FEE391").
    :type default_color: str, optional
    :param style: The style to use. Defaults to DEFAULT_STYLE ("perturbationx-default").
    :type style: str, optional
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """
    node_table = p4c.tables.get_table_columns(
        table="node", columns=[data_column], network=network_suid, base_url=cytoscape_url
    )
    data_range = [node_table.min()[data_column], node_table.max()[data_column]]

    p4c.style_mappings.set_node_color_mapping(
        data_column, data_range, colors=list(gradient), default_color=default_color,
        style_name=style, network=network_suid, base_url=cytoscape_url
    )


def highlight_subgraph(nodes: list, edges: list, network_suid: int, highlight_factor=3, cytoscape_url=DEFAULT_BASE_URL):
    """Highlight a subgraph.

    :param nodes: The nodes to highlight.
    :type nodes: list
    :param edges: The edges to highlight.
    :type edges: list
    :param network_suid: The network to update.
    :type network_suid: int
    :param highlight_factor: The factor to multiply the default width by. Defaults to 3.
    :type highlight_factor: float, optional
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """
    p4c.style_bypasses.set_node_border_width_bypass(
        nodes, DEFAULT_NODE_BORDER_WIDTH * highlight_factor,
        network=network_suid, base_url=cytoscape_url
    )
    p4c.style_bypasses.set_edge_line_width_bypass(
        edges, DEFAULT_EDGE_WIDTH * highlight_factor,
        network=network_suid, base_url=cytoscape_url
    )


def isolate_subgraph(graph: nx.Graph, nodes: list, edges: list, network_suid: int, cytoscape_url=DEFAULT_BASE_URL):
    """Isolate a subgraph.

    :param graph: The graph displayed in Cytoscape.
    :type graph: nx.Graph
    :param nodes: The nodes to isolate.
    :type nodes: list
    :param edges: The edges to isolate.
    :type edges: list
    :param network_suid: The network to update.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """
    # Set edge visibility
    core_edges = [edge_to_p4c_format(src, trg, graph[src][trg]["interaction"])
                  for src, trg in graph.edges if graph[src][trg]["type"] == "core"]
    edge_visibility = ["true" if e in edges else "false" for e in core_edges]
    p4c.style_bypasses.set_edge_property_bypass(
        core_edges, new_values=edge_visibility, visual_property="EDGE_VISIBLE",
        network=network_suid, base_url=cytoscape_url
    )

    # Set node visibility
    core_nodes = [n for n in graph.nodes if graph.nodes[n]["type"] == "core"]
    node_visibility = ["true" if n in nodes else "false" for n in core_nodes]
    p4c.style_bypasses.set_node_property_bypass(
        core_nodes, new_values=node_visibility, visual_property="NODE_VISIBLE",
        network=network_suid, base_url=cytoscape_url
    )


def extract_subgraph(nodes: list, edges: list, network_suid: int, cytoscape_url=DEFAULT_BASE_URL):
    """Extract a subgraph.

    :param nodes: The nodes to extract.
    :type nodes: list
    :param edges: The edges to extract.
    :type edges: list
    :param network_suid: The network to update.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    :return: The SUID of the extracted subnetwork.
    :rtype: int
    """
    return p4c.networks.create_subnetwork(
        nodes=nodes, nodes_by_col='name', edges=edges, edges_by_col='name',
        exclude_edges=True, network=network_suid, base_url=cytoscape_url
    )


def clear_bypass(components: list, component_type: str, visual_property: str, network_suid: int,
                 cytoscape_url=DEFAULT_BASE_URL):
    """Clear a bypass.

    :param components: The components to clear the bypass for.
    :type components: list
    :param component_type: The component type. Must be 'node' or 'edge'.
    :type component_type: str
    :param visual_property: The visual property to clear the bypass for.
    :type visual_property: str
    :param network_suid: The network to update.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    :raises ValueError: If the component type is not 'node' or 'edge'.
    :raises CyError: If a CyREST error occurs.
    """
    fake_logger = io.StringIO()
    with redirect_stderr(fake_logger):
        try:
            match component_type:
                case "node":
                    p4c.style_bypasses.clear_node_property_bypass(
                        components, visual_property, network_suid, base_url=cytoscape_url
                    )
                case "edge":
                    p4c.style_bypasses.clear_edge_property_bypass(
                        components, visual_property, network_suid, base_url=cytoscape_url
                    )
                case _:
                    raise ValueError("Invalid component type: '%s'. Must be 'node' or 'edge'." % component_type)
        except CyError as e:
            if "Bypass Visual Property does not exist" not in str(e):
                print(fake_logger.getvalue(), file=sys.stderr)
                raise e



from typing import Optional
import logging

import networkx as nx
import py4cytoscape as p4c
from py4cytoscape.py4cytoscape_utils import DEFAULT_BASE_URL

from perturbationx.result import NPAResult
from perturbationx.resources import DEFAULT_NODE_COLOR, DEFAULT_GRADIENT
from .cytoscape import set_boundary_display, edge_to_p4c_format, \
    color_nodes_by_column, highlight_subgraph, isolate_subgraph, extract_subgraph, clear_bypass


class NPAResultDisplay:
    """Class to display results from NPA analysis.

    :param graph: NetworkX graph object
    :type graph: nx.Graph
    :param results: NPA results object
    :type results: NPAResults
    :param network_style: Name of the network style to use
    :type network_style: str
    :param network_suid: SUID of the network to display results in.
    :type network_suid: int
    :param cytoscape_url: Cytoscape URL. Defaults to DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
    :type cytoscape_url: str, optional
    """

    def __init__(self, graph: nx.Graph, results: NPAResult, network_style: str, network_suid: int,
                 cytoscape_url=DEFAULT_BASE_URL):
        """Construct a new NPAResultDisplay object.
        """
        self._graph = graph
        self._results = results
        self._network_suid = network_suid
        self._network_style = network_style
        self._cytoscape_url = cytoscape_url

    def reset_display(self, display_boundary: Optional[bool] = None, reset_color=False,
                      reset_highlight=False, reset_visibility=False):
        """Reset the display of the network.

        :param display_boundary: Whether to display the boundary of the network. Defaults to None, which does not
            change the current setting.
        :type display_boundary: bool, optional
        :param reset_color: Whether to reset the color of the nodes. Defaults to False.
        :type reset_color: bool, optional
        :param reset_highlight: Whether to reset the highlight of the nodes and edges. Defaults to False.
        :type reset_highlight: bool, optional
        :param reset_visibility: Whether to reset the visibility of the nodes and edges. Defaults to False.
        :type reset_visibility: bool, optional
        :raises CyError: If a CyREST error occurs.
        :return: SUID of the network.
        :rtype: int
        """
        logging.getLogger().handlers.clear()  # Block logging to stdout

        p4c.set_current_network(self._network_suid, base_url=self._cytoscape_url)

        # Adjust boundary display as requested
        if display_boundary is not None:
            set_boundary_display(self._graph, display_boundary, self._network_suid, self._cytoscape_url)

        # Get core nodes and edges
        core_nodes = [n for n in self._graph.nodes if self._graph.nodes[n]["type"] == "core"]
        core_edges = [edge_to_p4c_format(src, trg, self._graph[src][trg]["interaction"])
                      for src, trg in self._graph.edges if self._graph[src][trg]["type"] == "core"]

        # Note: CyREST is very fussy with bypasses
        # you cannot remove a bypass if all given nodes/edges don't have it

        # Reset visibility
        if reset_visibility:
            isolate_subgraph(self._graph, core_nodes, core_edges,
                             network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)
            clear_bypass(core_edges, "edge", "EDGE_VISIBLE",
                         network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)
            clear_bypass(core_nodes, "node", "NODE_VISIBLE",
                         network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)

        # Reset highlight
        if reset_highlight:
            highlight_subgraph(core_nodes, core_edges, highlight_factor=1,
                               network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)
            clear_bypass(core_nodes, "node", "NODE_BORDER_WIDTH",
                         network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)
            clear_bypass(core_edges, "edge", "EDGE_WIDTH",
                         network_suid=self._network_suid, cytoscape_url=self._cytoscape_url)

        # Reset color
        if reset_color:
            p4c.style_mappings.delete_style_mapping(
                style_name=self._network_style, visual_prop="NODE_FILL_COLOR", base_url=self._cytoscape_url
            )

        return self._network_suid

    def color_nodes(self, dataset: str, attribute: str, gradient=DEFAULT_GRADIENT, default_color=DEFAULT_NODE_COLOR):
        """Color nodes by a given attribute.

        :param dataset: The dataset to color nodes by.
        :type dataset: str
        :param attribute: The attribute to color nodes by.
        :type attribute: str
        :param gradient: The gradient to use. Defaults to DEFAULT_GRADIENT ("#2B80EF", "#EF3B2C").
        :type gradient: (str, str), optional
        :param default_color: The default color to use. Defaults to DEFAULT_NODE_COLOR ("#FEE391").
        :type default_color: str, optional
        :raises CyError: If a CyREST error occurs.
        :return: SUID of the network.
        :rtype: int
        """
        self.reset_display(reset_color=True)

        data_column = dataset + ' ' + attribute
        color_nodes_by_column(data_column, self._network_suid, gradient, default_color,
                              self._network_style, self._cytoscape_url)

        return self._network_suid

    def highlight_leading_nodes(self, dataset: str, cutoff=0.8, attr="contribution", abs_value=True,
                                include_shortest_paths="none", path_length_tolerance=0,
                                include_neighbors=0, neighborhood_type="union"):
        """Highlight leading nodes.

        :param dataset: The dataset to highlight leading nodes for.
        :type dataset: str
        :param cutoff: The cutoff to use when determining leading nodes. Defaults to 0.8.
        :type cutoff: float, optional
        :param attr: The attribute to use when determining leading nodes. Defaults to "contribution".
        :type attr: str, optional
        :param abs_value: Whether to use the absolute value of the attribute. Defaults to True.
        :type abs_value: bool, optional
        :param include_shortest_paths: If "directed", the directed shortest paths between the nodes will be included.
            If "undirected", the undirected shortest paths between the nodes will be included. If "none",
            no shortest paths will be included. Defaults to "none".
        :type include_shortest_paths: str, optional
        :param path_length_tolerance: The tolerance for the length of the shortest paths. If 0, only the shortest paths
            are returned. If length_tolerance is an integer, it is interpreted as an absolute length. If
            length_tolerance is a float, it is interpreted as a percentage of the length of the shortest path.
            Defaults to 0.
        :type path_length_tolerance: int | float, optional
        :param include_neighbors: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
        :type include_neighbors: int, optional
        :param neighborhood_type: The type of neighborhood to include. Can be one of "union" or "intersection".
            If "union", all nodes within the maximum distance from any leading node are returned. If "intersection",
            only nodes within the maximum distance from all leading nodes are returned. Defaults to "union".
        :type neighborhood_type: str, optional
        :raises ValueError: If include_shortest_paths is not "directed", "undirected", or "none".
            If max_distance is less than 0 or neighborhood_type is not "union" or "intersection".
            If length_tolerance is not a number or is negative.
        :raises CyError: If a CyREST error occurs.
        :return: SUID of the network.
        :rtype: int
        """
        self.reset_display(reset_highlight=True)

        # Get subgraph
        leading_nodes = self._results.get_leading_nodes(dataset, cutoff=cutoff, attr=attr, abs_value=abs_value)
        nodes, edges = self._results.get_node_subgraph(
            leading_nodes, include_shortest_paths, path_length_tolerance, include_neighbors, neighborhood_type
        )
        edges = [edge_to_p4c_format(*e) for e in edges]

        # Pass nodes and edges to highlight function
        highlight_subgraph(
            nodes, edges,
            network_suid=self._network_suid,
            cytoscape_url=self._cytoscape_url
        )
        return self._network_suid

    def extract_leading_nodes(self, dataset: str, cutoff=0.8, attr="contribution", abs_value=True,
                              inplace=True, include_shortest_paths="none", path_length_tolerance=0,
                              include_neighbors=0, neighborhood_type="union"):
        """Extract leading nodes.

        :param dataset: The dataset to extract leading nodes for.
                :type dataset: str
        :param cutoff: The cutoff to use when determining leading nodes. Defaults to 0.8.
        :type cutoff: float, optional
        :param attr: The attribute to use when determining leading nodes. Defaults to "contribution".
        :type attr: str, optional
        :param abs_value: Whether to use the absolute value of the attribute. Defaults to True.
        :type abs_value: bool, optional
        :param inplace: Whether to extract the leading nodes in-place. Defaults to True. If True, the network will be
            modified by hiding all nodes and edges that are not leading nodes. If False, a new network will be created
            with only the leading nodes.
        :type inplace: bool, optional
        :param include_shortest_paths: If "directed", the directed shortest paths between the nodes will be included.
            If "undirected", the undirected shortest paths between the nodes will be included. If "none",
            no shortest paths will be included. Defaults to "none".
        :type include_shortest_paths: str, optional
        :param path_length_tolerance: The tolerance for the length of the shortest paths. If 0, only the shortest paths
            are returned. If length_tolerance is an integer, it is interpreted as an absolute length. If
            length_tolerance is a float, it is interpreted as a percentage of the length of the shortest path.
            Defaults to 0.
        :type path_length_tolerance: int | float, optional
        :param include_neighbors: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
        :type include_neighbors: int, optional
        :param neighborhood_type: The type of neighborhood to include. Can be one of "union" or "intersection".
            If "union", all nodes within the maximum distance from any leading node are returned. If "intersection",
            only nodes within the maximum distance from all leading nodes are returned. Defaults to "union".
        :type neighborhood_type: str, optional
        :raises ValueError: If include_shortest_paths is not "directed", "undirected", or "none".
            If max_distance is less than 0 or neighborhood_type is not "union" or "intersection".
            If length_tolerance is not a number or is negative.
        :raises CyError: If a CyREST error occurs.
        :return: SUID of the network.
        :rtype: int
        """
        self.reset_display(reset_visibility=True)

        # Get subgraph
        leading_nodes = self._results.get_leading_nodes(dataset, cutoff=cutoff, attr=attr, abs_value=abs_value)
        nodes, edges = self._results.get_node_subgraph(
            leading_nodes, include_shortest_paths, path_length_tolerance, include_neighbors, neighborhood_type
        )

        # Set visibility
        if inplace:
            edges = [edge_to_p4c_format(*e) for e in edges]
            isolate_subgraph(
                self._graph, nodes, edges,
                network_suid=self._network_suid,
                cytoscape_url=self._cytoscape_url
            )
            return self._network_suid
        else:
            edges = [edge_to_p4c_format(*e, escape_commas=True) for e in edges]
            return extract_subgraph(
                nodes, edges,
                network_suid=self._network_suid,
                cytoscape_url=self._cytoscape_url
            )

    def get_results(self):
        """Retrieve the results object for this display.

        :return: The results object.
        :rtype: NPAResults
        """
        return self._results



from .resources import *

__all__ = ['DEFAULT_DATA_COLS', 'DEFAULT_LOGGING_KWARGS', 'STYLE_XML', 'DEFAULT_STYLE', 'DEFAULT_NODE_COLOR',
           'DEFAULT_GRADIENT', 'DEFAULT_EDGE_WIDTH', 'DEFAULT_NODE_BORDER_WIDTH', 'get_style_xml_path',
           'get_osmotic_stress_core_path', 'get_osmotic_stress_boundary_path', 'get_osmotic_stress_dataset_paths']



from perturbationx.toponpa import permutation, preprocessing, statistics, core, matrices
from perturbationx.toponpa import toponpa as topo_npa

from .permutation import permute_adjacency, permute_edge_list
from .preprocessing import format_dataset, prune_network_dataset, infer_graph_attributes
from .statistics import compute_variances, confidence_interval, test_permutations, p_value
from .core import *
from .matrices import *
from .toponpa import *

__all__ = ['permutation', 'preprocessing', 'statistics', 'core', 'matrices', 'topo_npa', 'permute_adjacency',
           'permute_edge_list', 'format_dataset', 'prune_network_dataset', 'infer_graph_attributes',
           'compute_variances', 'confidence_interval', 'test_permutations', 'p_value', 'coefficient_inference',
           'perturbation_amplitude', 'perturbation_amplitude_contributions', 'generate_adjacencies',
           'generate_boundary_laplacian', 'generate_core_laplacians', 'toponpa', 'evaluate_modifications']



import warnings

import numpy as np
from scipy.stats import norm
from scipy.sparse import sparray

from .variance import core_covariance_matrix, perturbation_variance
from .permutation_tests import test_boundary_permutations, test_core_permutations

__all__ = ["compute_variances", "confidence_interval", "test_permutations", "p_value"]


def compute_variances(lap_b: np.ndarray | sparray, lap_c: np.ndarray | sparray, lap_q: np.ndarray | sparray,
                      stderr: np.ndarray, core_coefficients: np.ndarray, core_edge_count: int):
    """Compute the variance of the perturbation score and core node coefficients.

    :param lap_b: The Lb boundary Laplacian.
    :type lap_b: np.ndarray | sp.sparray
    :param lap_c: The Lc core Laplacian.
    :type lap_c: np.ndarray | sp.sparray
    :param lap_q: The Q core Laplacian.
    :type lap_q: np.ndarray | sp.sparray
    :param stderr: The standard error of the boundary coefficients.
    :type stderr: np.ndarray
    :param core_coefficients: The core node coefficients.
    :type core_coefficients: np.ndarray
    :param core_edge_count: The number of edges in the core network.
    :type core_edge_count: int
    :return: The variance of the perturbation score and core node coefficients.
    :rtype: (np.ndarray, np.ndarray)
    """
    core_covariance = core_covariance_matrix(lap_b, lap_c, stderr)
    node_variance = np.diag(core_covariance)
    npa_variance = perturbation_variance(lap_q, core_coefficients, core_covariance, core_edge_count)
    return npa_variance, node_variance


def confidence_interval(values: np.ndarray, variances: np.ndarray, alpha=0.95):
    """Compute the confidence intervals for the given significance level.

    :param values: The mean values for which to compute the confidence intervals.
    :type values: np.ndarray
    :param variances: The variances of the values.
    :type variances: np.ndarray
    :param alpha: The confidence level. Defaults to 0.95.
    :type alpha: float, optional
    :return: The lower and upper confidence intervals and the p-values.
    :rtype: (np.ndarray, np.ndarray, np.ndarray)
    """

    percentile_threshold = norm.ppf((1. + alpha) / 2.)
    std_dev = np.sqrt(variances)

    ci_lower = values - percentile_threshold * std_dev
    ci_upper = values + percentile_threshold * std_dev
    p_values = np.subtract(1., norm.cdf(np.divide(np.abs(values), std_dev)))
    return ci_lower, ci_upper, p_values


def test_permutations(lap_b: np.ndarray | sparray, lap_c: np.ndarray | sparray, lap_q: np.ndarray | sparray,
                      adj_perms: dict, core_edge_count: int, boundary_coefficients: np.ndarray,
                      permutations=('o', 'k2'), full_core_permutation=True, exact_boundary_outdegree=True,
                      permutation_rate=1., iterations=500, seed=None):
    """Test the null hypothesis that the ordering of boundary coefficients and the distribution of core edges does not
    affect the perturbation score. This is a convenience function that calls test_boundary_permutations and
    test_core_permutations.

    :param lap_b: The Lb boundary Laplacian.
    :type lap_b: np.ndarray | sp.sparray
    :param lap_c: The Lc core Laplacian.
    :type lap_c: np.ndarray | sp.sparray
    :param lap_q: The Q core Laplacian.
    :type lap_q: np.ndarray | sp.sparray
    :param adj_perms: The adjacency matrices of the core network permutations. The keys are the permutation names,
                        the values are the adjacency matrices. The adjacency matrices may be sparse or dense.
    :type adj_perms: dict
    :param core_edge_count: The number of edges in the core network.
    :type core_edge_count: int
    :param boundary_coefficients: The boundary coefficients.
    :type boundary_coefficients: np.ndarray
    :param permutations: The permutations to test. May contain 'o', 'k1', and 'k2' in any order.
                            Defaults to ('o', 'k2'). For 'k1' and 'k2', the adjacency matrices to test must be
                            provided in adj_perms.
    :type permutations: list, optional
    :param full_core_permutation: Whether to use the full permutation matrix for each core permutation. Partial
                                    permutations sample core coefficients, while full permutations sample perturbation
                                    scores. Defaults to True.
    :type full_core_permutation: bool, optional
    :param exact_boundary_outdegree: Whether to use the exact boundary outdegree. If False, the boundary outdegree
                                        is set to 1 for all core nodes with boundary edges. Defaults to True.
    :type exact_boundary_outdegree: bool, optional
    :param permutation_rate: The fraction of boundary coefficients to permute. Defaults to 1.
    :type permutation_rate: float, optional
    :param iterations: The number of boundary permutations to perform. Defaults to 500.
    :type iterations: int, optional
    :param seed: The seed for the random number generator. Defaults to None.
    :type seed: int, optional
    :return: The distributions of perturbation scores under the null hypothesis.
    :rtype: dict
    """

    distributions = dict()
    for p in set(permutations):
        match p.lower():
            case 'o':
                distributions[p] = test_boundary_permutations(
                    lap_b=lap_b, lap_c=lap_c, lap_q=lap_q,
                    boundary_coefficients=boundary_coefficients,
                    core_edge_count=core_edge_count, permutation_rate=permutation_rate,
                    iterations=iterations, seed=seed
                )
            case 'k1' | 'k2':
                distributions[p] = test_core_permutations(
                    adj_perms=adj_perms[p], boundary_coefficients=boundary_coefficients,
                    lap_b=lap_b, core_edge_count=core_edge_count,
                    exact_boundary_outdegree=exact_boundary_outdegree,
                    lap_q=lap_q, full_permutations=full_core_permutation
                )
            case _:
                warnings.warn("Permutation %s is unknown and will be skipped." % p)
                continue

    return distributions


def p_value(value: float, distribution: list):
    """Compute the p-value for the given value in the given distribution.

    :param value: The value for which to compute the p-value.
    :type value: float
    :param distribution: The distribution.
    :type distribution: list
    :return: The p-value.
    :rtype: float
    """
    return sum(sample > value for sample in distribution) / len(distribution)



import platform
import copy
import warnings
import logging
from datetime import datetime
import json
from importlib.metadata import version, PackageNotFoundError

import matplotlib.pyplot as plt
import pandas
import seaborn as sns
import networkx as nx
import py4cytoscape as p4c
from py4cytoscape.py4cytoscape_utils import DEFAULT_BASE_URL

from perturbationx.resources import DEFAULT_STYLE
from perturbationx.vis import NPAResultDisplay
from perturbationx.vis.cytoscape import init_cytoscape, load_network_data
from perturbationx.util import get_shortest_path_components, get_neighborhood_components


class NPAResult:
    """Class for storing and accessing the results of a Network Perturbation Analysis (NPA). It is recommended
    to build an NPAResult object using NPAResultBuilder to ensure correct formatting. Metadata is prefixed with
    "network\_" to avoid conflicts, unless the metadata key already starts with "network" or "dataset". By default,
    the following metadata is added: datetime_utc, python_implementation, python_version, system_name,
    system_release, system_version, network_title, network_collection, perturbationx_version, numpy_version,
    networkx_version, pandas_version, scipy_version, matplotlib_version, seaborn_version, and py4cytoscape_version.


    :param graph: The network graph.
    :type graph: networkx.DiGraph
    :param datasets: The datasets used for the analysis.
    :type datasets: list
    :param global_info: The global information for each dataset.
    :type global_info: pandas.DataFrame
    :param node_info: The node information for each dataset.
    :type node_info: pandas.DataFrame
    :param distributions: The distributions for each permutation.
    :type distributions: dict
    :param metadata: Additional metadata to store with the result.
    :type metadata: dict, optional
    """
    def __init__(self, graph: nx.DiGraph, datasets: list, global_info: pandas.DataFrame, node_info: pandas.DataFrame,
                 distributions: dict, metadata=None):
        """Construct a new NPAResult.
        """

        self._graph = graph
        self._datasets = datasets
        self._global_info = global_info
        self._node_info = node_info
        self._distributions = distributions
        self._metadata = metadata if metadata is not None else dict()

        # System metadata
        self._metadata["datetime_utc"] = datetime.utcnow().isoformat()
        self._metadata["python_implementation"] = platform.python_implementation()
        self._metadata["python_version"] = platform.python_version()
        self._metadata["system_name"] = platform.system()
        self._metadata["system_release"] = platform.release()
        self._metadata["system_version"] = platform.version()

        # Network metadata
        for k, v in graph.graph.items():
            if k.startswith("dataset") or k.startswith("network"):
                self._metadata[k] = v
            self._metadata["network_" + k] = v

        if "network_title" not in self._metadata:
            self._metadata["network_title"] = "Untitled network"
        if "network_collection" not in self._metadata:
            self._metadata["network_collection"] = "Untitled collection"

        # Library metadata
        try:
            self._metadata["perturbationx_version"] = version("perturbationx")
        except PackageNotFoundError:
            self._metadata["perturbationx_version"] = "dev"

        self._metadata["numpy_version"] = version("numpy")
        self._metadata["networkx_version"] = version("networkx")
        self._metadata["pandas_version"] = version("pandas")
        self._metadata["scipy_version"] = version("scipy")
        self._metadata["matplotlib_version"] = version("matplotlib")
        self._metadata["seaborn_version"] = version("seaborn")
        self._metadata["py4cytoscape_version"] = version("py4cytoscape")

    def metadata(self):
        """Get the metadata for this result.

        :return: The metadata for this result.
        :rtype: dict
        """
        return copy.deepcopy(self._metadata)

    def datasets(self):
        """Get the datasets used for this result.

        :return: The datasets used for this result.
        :rtype: list
        """
        return copy.deepcopy(self._datasets)

    def node_attributes(self):
        """Get the node attributes for this result.

        :return: The node attributes for this result.
        :rtype: list
        """
        return self._node_info.index.unique('attr').tolist()

    def distributions(self):
        """Get the distributions for this result.

        :return: The distributions for this result.
        :rtype: list
        """
        return list(self._distributions.keys())

    def global_info(self):
        """Get the global information for this result.

        :return: The global information for this result.
        :rtype: pandas.DataFrame
        """
        return self._global_info.copy()

    def node_info(self, accessor: str):
        """Get the node information for this result.

        :param accessor: The dataset or node attribute to get the information for.
        :type accessor: str
        :return: The node information for this result.
        :rtype: pandas.DataFrame
        """
        level = "data" if accessor in self._datasets else "attr"
        return self._node_info.xs(accessor, level=level).transpose(copy=True)

    def get_distribution(self, distribution: str, dataset: str, include_reference=False):
        """Get the distribution for a permutation.

        :param distribution: The permutation to get the distribution for.
        :type distribution: str
        :param dataset: The dataset to get the distribution for.
        :type dataset: str
        :param include_reference: If True, the reference value will be included in the distribution. Defaults to False.
        :type include_reference: bool, optional
        :return: The distribution for the permutation. If include_reference is True, a tuple of the distribution and
            the reference value will be returned.
        :rtype: list | (list, float)
        """
        value_distribution = copy.deepcopy(self._distributions[distribution][dataset][0])
        if include_reference:
            return value_distribution, self._distributions[distribution][dataset][1]
        else:
            return value_distribution

    def plot_distribution(self, distribution: str, datasets=None, show=True):
        """Plot the distribution for a permutation.

        :param distribution: The permutation to plot the distribution for.
        :type distribution: str
        :param datasets: The datasets to plot the distribution for. If None, all datasets will be plotted.
        :type datasets: list, optional
        :param show: If True, the plot will be shown. Defaults to True.
        :type show: bool, optional
        :return: The axes of the plot.
        :rtype: matplotlib.axes.Axes
        """
        if datasets is None:
            datasets = self._datasets
        else:
            datasets = [d for d in datasets if d in self._datasets]

        if len(datasets) == 0:
            warnings.warn('Nothing to plot for distribution %s and datasets %s.' % (distribution, str(datasets)))
            return

        plt.clf()
        fig, ax = plt.subplots(nrows=len(datasets), figsize=(8, 4*len(datasets)))
        if len(datasets) == 1:
            ax = [ax]
        plt.suptitle("Permutation '%s'" % distribution)

        for idx, d in enumerate(datasets):
            distr = self._distributions[distribution][d]
            ax[idx].set_ylabel(d)

            sns.histplot(distr[0], ax=ax[idx], bins=25, stat="density")
            sns.kdeplot(distr[0], ax=ax[idx], color="tab:blue")
            if distr[1] is not None:
                ax[idx].axvline(x=distr[1], color="tab:red")

        plt.tight_layout(rect=[0.01, 0.01, 1, 0.97])
        if show:
            plt.show()
        return ax

    def get_leading_nodes(self, dataset: str, cutoff=0.8, attr="contribution", abs_value=True):
        """Get the leading nodes for a dataset. The leading nodes are the nodes that contribute the most
        to a selected attribute, up to a certain cutoff.

        :param dataset: The dataset to get the leading nodes for.
        :type dataset: str
        :param cutoff: The cutoff for the cumulative distribution. Defaults to 0.8.
        :type cutoff: float, optional
        :param attr: The node attribute to get the leading nodes for. Defaults to "contribution".
        :type attr: str, optional
        :param abs_value: If True, the absolute value of the attribute will be used. Defaults to True.
        :type abs_value: bool, optional
        :return: The leading nodes for the dataset.
        :rtype: set
        """
        contributions = self.node_info(attr)[dataset]
        if abs_value:
            contributions = contributions.abs()
        contributions = contributions.sort_values(ascending=False)

        cumulative_contributions = contributions.cumsum() / contributions.sum()
        max_idx = 0
        for contr in cumulative_contributions:
            max_idx += 1
            if contr > cutoff:
                break
        return set(cumulative_contributions.index[:max_idx].tolist())

    def get_node_subgraph(self, nodes, include_shortest_paths="none", path_length_tolerance=0,
                          include_neighbors=0, neighborhood_type="union"):
        """Get the subgraph for a set of nodes. The subgraph can include the shortest paths between the nodes,
        the neighborhood of the nodes, or both.

        :param nodes: The nodes to get the subgraph for.
        :type nodes: set
        :param include_shortest_paths: If "directed", the directed shortest paths between the nodes will be included.
            If "undirected", the undirected shortest paths between the nodes will be included. If "none",
            no shortest paths will be included. Defaults to "none".
        :type include_shortest_paths: str, optional
        :param path_length_tolerance: The tolerance for the length of the shortest paths. If 0, only the shortest paths
            are returned. If length_tolerance is an integer, it is interpreted as an absolute length. If
            length_tolerance is a float, it is interpreted as a percentage of the length of the shortest path.
            Defaults to 0.
        :type path_length_tolerance: int | float, optional
        :param include_neighbors: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
        :type include_neighbors: int, optional
        :param neighborhood_type: The type of neighborhood to include. Can be one of "union" or "intersection".
            If "union", all nodes within the maximum distance from any leading node are returned. If "intersection",
            only nodes within the maximum distance from all leading nodes are returned. Defaults to "union".
        :type neighborhood_type: str, optional
        :raises ValueError: If include_shortest_paths is not "directed", "undirected", or "none".
            If max_distance is less than 0 or neighborhood_type is not "union" or "intersection".
            If length_tolerance is not a number or is negative.
        :return: The nodes and edges in the subgraph. They are returned as a pair of lists.
        :rtype: (list, list)
        """
        # Remove boundary nodes
        graph = self._graph.copy()
        boundary_nodes = [n for n in graph.nodes if graph.nodes[n]["type"] == "boundary"]
        graph.remove_nodes_from(boundary_nodes)

        # Find paths
        match include_shortest_paths:
            case "directed":
                path_nodes, path_edges = get_shortest_path_components(
                    graph, nodes, directed=True, length_tolerance=path_length_tolerance
                )
            case "undirected":
                path_nodes, path_edges = get_shortest_path_components(
                    graph, nodes, directed=False, length_tolerance=path_length_tolerance
                )
            case "none" | None:
                path_nodes, path_edges = set(), set()
            case _:
                raise ValueError("Argument include_shortest_paths must be "
                                 "'directed', 'undirected', or 'none'.")

        # Find neighborhood
        neigh_nodes, neigh_edges = get_neighborhood_components(graph, nodes, include_neighbors, neighborhood_type)

        nodes = list(nodes.union(path_nodes).union(neigh_nodes))
        edges = list(path_edges.union(neigh_edges))
        return nodes, edges

    def display_network(self, display_boundary=False, style=DEFAULT_STYLE, cytoscape_url=DEFAULT_BASE_URL):
        """Display the network in Cytoscape.

        :param display_boundary: If True, boundary nodes will be displayed. Defaults to False.
        :type display_boundary: bool, optional
        :param style: The style to apply to the network. Defaults to DEFAULT_STYLE ("perturbationx-default").
        :type style: str, optional
        :param cytoscape_url: The URL of the Cytoscape instance to display the network in. Defaults to
            DEFAULT_BASE_URL in py4cytoscape (http://127.0.0.1:1234/v1).
        :type cytoscape_url: str, optional
        :return: The display object.
        :rtype: perturbationx.NPAResultDisplay
        """
        logging.getLogger().handlers.clear()  # Block logging to stdout

        # Initialize Cytoscape
        network_suid = init_cytoscape(
            self._graph, self._metadata['network_title'], self._metadata['network_collection'],
            display_boundary, network_suid=None, cytoscape_url=cytoscape_url
        )

        # Apply style
        p4c.styles.set_visual_style(style, network=network_suid, base_url=cytoscape_url)

        # Load node data
        load_network_data(self._node_info.transpose(), "node", network_suid, cytoscape_url)

        return NPAResultDisplay(self._graph, self, style, network_suid, cytoscape_url)

    def to_networkx(self):
        """Retrieve the NetworkX graph for this result.

        :return: The NetworkX graph.
        :rtype: networkx.DiGraph
        """
        graph_copy = self._graph.copy()
        graph_copy.graph.update(self._metadata)
        for n in graph_copy.nodes:
            graph_copy.nodes[n].update(self._node_info.loc[(n, "data")].to_dict())

        node_dataframe = self._node_info.transpose(copy=True)
        # Rename MultiIndex columns
        node_dataframe.columns = [col if isinstance(col, str) else ' '.join(col)
                                  for col in node_dataframe.columns]
        node_attr = node_dataframe.to_dict('index')
        nx.set_node_attributes(graph_copy, node_attr)

        return graph_copy

    def to_dict(self):
        """Convert this result to a dictionary.

        :return: The result as a dictionary. Top-level keys are "metadata" and dataset names. For each dataset, the
            top-level keys are "global_info", "node_info", and "distributions".
        :rtype: dict
        """
        data = dict()
        data["metadata"] = self._metadata.copy()

        for d in self.datasets():
            data[d] = dict()

            data[d]["global_info"] = dict()
            for column in self._global_info.columns:
                data[d]["global_info"][column] = self._global_info.loc[d, column]

            data[d]["node_info"] = dict()
            for attr in self.node_attributes():
                data[d]["node_info"][attr] = {n: self._node_info.loc[(d, attr), n] for n in self._node_info.columns}

            data[d]["distributions"] = dict()
            for distr in self.distributions():
                data[d]["distributions"][distr] = self._distributions[distr][d][0]

        return data

    def to_json(self, filepath: str, indent=4):
        """Save this result to a JSON file. The format is the same as the output of to_dict().

        :param filepath: The path to save the result to.
        :type filepath: str
        :param indent: The indentation to use. Defaults to 4.
        :type indent: int, optional
        """
        data = self.to_dict()
        with open(filepath, 'w') as f:
            json.dump(data, f, ensure_ascii=False, indent=indent)



import networkx as nx

__all__ = ["get_neighborhood_components"]


def get_all_neighbors(graph: nx.Graph, source_nodes: set, max_distance=0):
    """Get all neighbors of a set of nodes in a graph.

    :param graph: The graph to search for neighbors in.
    :type graph: nx.Graph
    :param source_nodes: The nodes to find neighbors of.
    :type source_nodes: set
    :param max_distance: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
    :type max_distance: int, optional
    :return: The neighbors of the source nodes.
    :rtype: set
    """
    if max_distance < 1:
        return source_nodes

    neighbors = set()
    for n in source_nodes:
        for m in graph.neighbors(n):
            if m not in source_nodes:
                neighbors.add(m)

    return get_all_neighbors(graph, neighbors, max_distance - 1).union(source_nodes)


def get_common_neighbors(graph: nx.Graph, source_nodes: set, max_distance=0):
    """Get the common neighbors of a set of nodes in a graph.

    :param graph: The graph to search for neighbors in.
    :type graph: nx.Graph
    :param source_nodes: The nodes to find neighbors of.
    :type source_nodes: set
    :param max_distance: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
    :type max_distance: int, optional
    :return: The common neighbors of the source nodes.
    :rtype: set
    """

    neighborhoods = {n: get_all_neighbors(graph, {n}, max_distance) for n in source_nodes}

    neighbors = None
    for n in neighborhoods:
        if neighbors is None:
            neighbors = neighborhoods[n]
        else:
            neighbors = neighbors.intersection(neighborhoods[n])

    return neighbors


def get_neighborhood_components(graph: nx.DiGraph, source_nodes: set, max_distance=0, neighborhood_type="union"):
    """Get the neighborhood of a set of nodes in a graph.

    :param graph: The graph to search for neighbors in.
    :type graph: nx.DiGraph
    :param source_nodes: The nodes to find neighbors of.
    :type source_nodes: set
    :param max_distance: The maximum distance from leading nodes that neighbors can be. If 0, no neighbors
            will be included. If 1, only the neighbors of the nodes will be included. If 2, the neighbors of the
            neighbors of the nodes will be included, and so on. Defaults to 0.
    :type max_distance: int, optional
    :param neighborhood_type: The type of neighborhood to find. Can be one of "union" or "intersection".
            If "union", all nodes within the maximum distance from any leading node are returned. If "intersection",
            only nodes within the maximum distance from all leading nodes are returned. Defaults to "union".
    :type neighborhood_type: str, optional
    :raises ValueError: If max_distance is less than 0 or neighborhood_type is not "union" or "intersection".
    :return: The nodes and edges in the neighborhood. They are returned as a pair of sets.
    :rtype: (set, set)
    """

    if max_distance < 0.:
        raise ValueError("Argument max_distance must be >= 0.")
    elif max_distance == 0:
        return source_nodes, set()

    match neighborhood_type:
        case "union":
            nodes = get_all_neighbors(graph.to_undirected(), source_nodes, max_distance)
        case "intersection":
            nodes = get_common_neighbors(graph.to_undirected(), source_nodes, max_distance)
        case _:
            raise ValueError("Argument neighborhood_type must be 'union' or 'intersection'.")

    edges = set()
    for src, trg in graph.edges:
        if src in nodes and trg in nodes:
            edges.add((src, trg, graph[src][trg]["interaction"]))

    return nodes, edges
